from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math

def init_Octahedron(cls):
    cls.edge_length = math.sqrt(2)
    cls.base_positions = \
    [
        glm.vec3(-1,  0,  0), # 0
        glm.vec3( 1,  0,  0), # 1

        glm.vec3( 0, -1,  0), # 2
        glm.vec3( 0,  1,  0), # 3

        glm.vec3( 0,  0, -1), # 4
        glm.vec3( 0,  0,  1), # 5
    ]

    cls.base_indices = \
    [
        glm.uvec3(5, 0, 2),
        glm.uvec3(5, 2, 1),
        glm.uvec3(5, 1, 3),
        glm.uvec3(5, 3, 0),

        glm.uvec3(4, 2, 0),
        glm.uvec3(4, 1, 2),
        glm.uvec3(4, 3, 1),
        glm.uvec3(4, 0, 3)
    ]

    cls.fixed_vertices = []
    for index in cls.base_indices:
        pos0 = cls.base_positions[index[0]]
        pos1 = cls.base_positions[index[1]]
        pos2 = cls.base_positions[index[2]]

        v1 = pos1 - pos0
        v2 = pos2 - pos0
        normal = glm.normalize(glm.cross(v1, v2))

        vertex0 = Vertex(
            position=pos0,
            normal=normal,
            tex_coord=glm.vec3(0.5, 1, 0)
        )
        vertex1 = Vertex(
            position=pos1,
            normal=normal,
            tex_coord=glm.vec3(0.5-math.sqrt(3)/4, 0.25, 0)
        )
        vertex2 = Vertex(
            position=pos2,
            normal=normal,
            tex_coord=glm.vec3(0.5+math.sqrt(3)/4, 0.25, 0)
        )
        cls.fixed_vertices.append(vertex0)
        cls.fixed_vertices.append(vertex1)
        cls.fixed_vertices.append(vertex2)

    return cls

@init_Octahedron
class Octahedron(Mesh):

    @checktype
    def __init__(self, radius=1,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 stable=False, normalize_tex_coord=False,
                 name:str=""):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=True)
        self.__radius = radius
        self.__stable = stable
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = True
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices
        radius = self.__radius
        stable = self.__stable
        normalize_tex_coord = self.__normalize_tex_coord

        quat = glm.quat(1, 0, 0, 0)
        if stable:
            quat_z = glm.quat(math.cos(math.pi/8), 0, 0, math.sin(math.pi/8))

            theta = math.acos(math.sqrt(3)/3)
            quat_x = glm.quat(math.cos(theta/2), math.sin(theta/2), 0, 0)

            quat = quat_x * quat_z

        for i, fix_vert in enumerate(Octahedron.fixed_vertices):
            tex_coord = fix_vert.tex_coord
            if not normalize_tex_coord:
                tex_coord = 2*radius*(tex_coord - glm.vec3(0.5, 0.5, 0))*Octahedron.edge_length/math.sqrt(3) + glm.vec3(0.5, 0.5, 0)
            
            vertices[i] = Vertex(
                position = radius * (quat * fix_vert.position),
                normal = quat * fix_vert.normal,
                tex_coord = tex_coord
            )

        for i in range(0, len(Octahedron.fixed_vertices), 3):
            indices[int(i/3)] = glm.uvec3(i, i+1, i+2)
            self.generate_temp_TBN(vertices[i], vertices[i+1], vertices[i+2])

    @property
    def radius(self):
        return self.__radius
    
    @radius.setter
    @Mesh.param_setter
    def radius(self, radius:float):
        self.__radius = radius

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord
    
    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag

    @property
    def is_stable(self):
        return self.__stable
    
    @is_stable.setter
    @Mesh.param_setter
    def is_stable(self, flag:bool):
        self.__stable = flag


from .SceneNode import SceneNode
from .Mesh import Mesh
from .Lights.PointLight import PointLights, PointLight, FlatPointLight
from .Lights.DirLight import DirLights, DirLight, FlatDirLight
from .Lights.SpotLight import SpotLights, SpotLight, FlatSpotLight
from .AffineTransform import AffineTransform

from .Fog import Fog
from .Background import Background

from glass.utils import quat_to_mat4, scale_to_mat4, translate_to_mat4
from glass import Instances

import glm
import time

class Scene:

    def __init__(self):
        self._root = SceneNode("root")
        self._root._scenes.add(self)

        self._all_meshes = {}
        self._backup_meshes = {}
        self._last_generated_meshes = set()

        self._fog = Fog()
        self._background = Background()
        self._dir_lights = DirLights()
        self._point_lights = PointLights()
        self._spot_lights = SpotLights()

        self.__anything_changed = False
        self._should_update_env_maps = False
        self._should_update_depth_maps = False

    def add(self, scene_node:SceneNode)->None:
        self._root.add_child(scene_node)

    def remove(self, scene_node:SceneNode)->None:
        self._root.remove_child(scene_node)

    @property
    def root(self):
        return self._root

    @property
    def background(self)->Background:
        return self._background

    @property
    def skybox(self):
        return self._background.skybox
    
    @skybox.setter
    def skybox(self, skybox):
        self._background.skybox = skybox

    @property
    def skydome(self):
        return self._background.skydome
    
    @skydome.setter
    def skydome(self, skydome):
        self._background.skydome = skydome

    @property
    def fog(self):
        return self._fog

    def __update_env_maps(self, scene_node:SceneNode=None):
        if scene_node is None:
            scene_node = self._root

        if isinstance(scene_node, Mesh) and scene_node.material.auto_update_env_map:
            scene_node.material.update_env_map()

        for child in scene_node.children:
            self.__update_env_maps(child)

    def __update_depth_maps(self):
        success = False
        for point_light in self._point_lights:
            if not point_light.need_update_depth_map:
                point_light.need_update_depth_map = True
                success = True

        if success:
            self._point_lights.dirty = True

        success = False
        for spot_light in self._spot_lights:
            if not spot_light.need_update_depth_map:
                spot_light.need_update_depth_map = True
                success = True

        if success:
            self._spot_lights.dirty = True

    def __clear_dirty(self, scene_node:SceneNode=None):
        if scene_node is None:
            scene_node = self._root

        if self not in scene_node._transform_dirty and \
           self not in scene_node._children_transform_dirty:
            return
        
        if self in scene_node._transform_dirty:
            scene_node._transform_dirty.remove(self)

        if self in scene_node._children_transform_dirty:
            for child in scene_node._children_transform_dirty[self]:
                self.__clear_dirty(child)
            del scene_node._children_transform_dirty[self]
    
    def __trav(self, scene_node:SceneNode,
               current_quat:glm.quat,
               current_mat:glm.mat4,
               current_path:str):
        
        if self not in self._root._transform_dirty and \
           self not in self._root._children_transform_dirty:
            return

        new_quat = current_quat * scene_node.orientation
        new_mat = current_mat * translate_to_mat4(scene_node.position) * \
                                quat_to_mat4(scene_node.orientation) * \
                                scale_to_mat4(scene_node.scale)
        new_path = current_path + "/" + scene_node.name

        if self in scene_node._transform_dirty:
            if isinstance(scene_node, Mesh):
                mesh = scene_node
                if mesh not in self._all_meshes:
                    self._all_meshes[mesh] = {}

                if new_path not in self._all_meshes[mesh]:
                    self._all_meshes[mesh][new_path] = AffineTransform()

                self._all_meshes[mesh][new_path]["affine_transform_row0"] = glm.vec4(new_mat[0][0], new_mat[1][0], new_mat[2][0], new_mat[3][0])
                self._all_meshes[mesh][new_path]["affine_transform_row1"] = glm.vec4(new_mat[0][1], new_mat[1][1], new_mat[2][1], new_mat[3][1])
                self._all_meshes[mesh][new_path]["affine_transform_row2"] = glm.vec4(new_mat[0][2], new_mat[1][2], new_mat[2][2], new_mat[3][2])
                self.__anything_changed = True
            elif isinstance(scene_node, SpotLight):
                spot_light = None
                if new_path not in self._spot_lights:
                    spot_light = FlatSpotLight(scene_node)
                    self._spot_lights[new_path] = spot_light
                else:
                    spot_light = self._spot_lights[new_path]
                    spot_light.update(scene_node)

                spot_light.abs_position = new_mat[3].xyz
                spot_light.direction = new_quat * glm.vec3(0, 1, 0)
                self._spot_lights.dirty = True

                self.__anything_changed = True
                self._spot_lights_changed = True
            elif isinstance(scene_node, PointLight):
                point_light = None
                if new_path not in self._point_lights:
                    point_light = FlatPointLight(scene_node)
                    self._point_lights[new_path] = point_light
                else:
                    point_light = self._point_lights[new_path]
                    point_light.update(scene_node)

                point_light.abs_position = new_mat[3].xyz
                self._point_lights.dirty = True

                self.__anything_changed = True
                self._point_lights_changed = True
            elif isinstance(scene_node, DirLight):
                dir_light = None
                if new_path not in self._dir_lights:
                    dir_light = FlatDirLight(scene_node)
                    self._dir_lights[new_path] = dir_light
                else:
                    dir_light = self._dir_lights[new_path]
                    dir_light.update(scene_node)

                dir_light.direction = new_quat * glm.vec3(0, 1, 0)
                dir_light.abs_orientation = new_quat
                self._dir_lights.dirty = True

                self.__anything_changed = True
                self._dir_lights_changed = True

        if self in scene_node._children_transform_dirty:
            for child in scene_node._children_transform_dirty[self]:
                self.__trav(child, new_quat, new_mat, new_path)

    def __collect_render_infos(self):
        if self not in self._root._transform_dirty and \
           self not in self._root._children_transform_dirty:
            return
        
        self.__trav(self._root, glm.quat(), glm.mat4(), "")
        if self.__anything_changed:
            self.__update_env_maps()
            self.__update_depth_maps()
            self.__anything_changed = False
            self._should_update_env_maps = False
            self._should_update_depth_maps = False
        else:
            if self._should_update_env_maps:
                self.__update_env_maps()
                self._should_update_env_maps = False

            if self._should_update_depth_maps:
                self.__update_depth_maps()
                self._should_update_depth_maps = False

        self.__clear_dirty()
        for mesh in self._all_meshes:
            if mesh not in self._backup_meshes:
                assert isinstance(self._all_meshes[mesh], dict)
                self._backup_meshes[mesh] = Instances(self._all_meshes[mesh])
                self._all_meshes[mesh] = self._backup_meshes[mesh]
            elif isinstance(self._all_meshes[mesh], dict):
                self._backup_meshes[mesh].update(self._all_meshes[mesh])
                self._all_meshes[mesh] = self._backup_meshes[mesh]

    @property
    def dir_lights(self):
        self.__collect_render_infos()
        return self._dir_lights
    
    @property
    def point_lights(self):
        self.__collect_render_infos()
        return self._point_lights
    
    @property
    def spot_lights(self):
        self.__collect_render_infos()
        return self._spot_lights
    
    @property
    def all_meshes(self):
        self.__collect_render_infos()
        return self._all_meshes
    
    def generate_meshes(self, seconds:float=0.01):
        generating_meshes = set()
        for mesh in self.all_meshes.keys():
            if mesh.is_generating:
                generating_meshes.add(mesh)
        
        if not generating_meshes:
            return False

        if self._last_generated_meshes >= generating_meshes:
            self._last_generated_meshes.clear()

        start_time = time.time()
        overtime = False
        while True:
            for mesh in (generating_meshes - self._last_generated_meshes):
                mesh.generate()
                current_time = time.time()
                self._last_generated_meshes.add(mesh)

                if current_time - start_time > seconds:
                    overtime = True
                    break

            if overtime:
                break
            
            self._last_generated_meshes.clear()
        
        self.__anything_changed = True
        return True
    
    @staticmethod
    def __remove_path_prefix(instance_map, path_str):
        if isinstance(instance_map, dict):
            should_remove_meshes = []
            for mesh, instances in instance_map.items():
                should_remove_keys = []
                for key in instances.keys():
                    if key == path_str or key.starts_with(path_str + "/"):
                        should_remove_keys.append(key)
                
                if len(should_remove_keys) == len(instances):
                    should_remove_meshes.append(mesh)
                else:
                    for key in should_remove_keys:
                        del instances[key]

            for mesh in should_remove_meshes:
                del instance_map[mesh]
        else:
            lights = instance_map
            should_remove_keys = []
            for key in lights.keys():
                if key == path_str or key.starts_with(path_str + "/"):
                    should_remove_keys.append(key)

            for key in should_remove_keys:
                lights[key].before_del()
                del lights[key]

    def _remove_paths_prefix(self, paths_str:set):
        for path_str in paths_str:
            Scene.__remove_path_prefix(self._all_meshes, path_str)
            Scene.__remove_path_prefix(self._dir_lights, path_str)
            Scene.__remove_path_prefix(self._point_lights, path_str)
            Scene.__remove_path_prefix(self._spot_lights, path_str)



import copy
from OpenGL import GL
import glm
import numpy as np

from .GLConfig import GLConfig
from .VAO import VAO
from .utils import checktype, di
from .AttrList import AttrList
from .GLInfo import GLInfo
from .SameTypeList import SameTypeList

class Vertex(dict):
    _all_attrs = \
    {
        '__class__', '__contains__', '__delattr__', '__delitem__',
        '__dict__', '__dir__', '__doc__', '__eq__', '__format__',
        '__ge__', '__getattribute__', '__getitem__', '__gt__',
        '__hash__', '__init__', '__init_subclass__', '__iter__',
        '__le__', '__len__', '__lt__', '__module__', '__ne__', '__new__',
        '__reduce__', '__reduce_ex__', '__repr__', '__setattr__',
        '__setitem__', '__sizeof__', '__str__', '__subclasshook__',
        '__weakref__', '__getattr__', '__setattr__', '__deepcopy__', '__setstate__',
        'clear', 'copy', 'fromkeys', 'get', 'items', 'keys',
        'pop', 'popitem', 'setdefault', 'update', 'values',
        '_array_index_map', '_add_array_index', "user_data"
    }

    def __init__(self, **kwargs):
        self._array_index_map = {}
        self.user_data = {}
        dict.__init__(self, **kwargs)

    def _add_array_index(self, array, index):
        id_array = id(array)
        if id_array not in self._array_index_map:
            self._array_index_map[id_array] = set()

        self._array_index_map[id_array].add(index)

    def __hash__(self):
        return id(self)

    def __deepcopy__(self, memo):
        vertex = self.__class__()
        for key in self.keys():
            vertex[key] = copy.deepcopy(self[key], memo)
        return vertex

    def __getattr__(self, name):
        if name in Vertex._all_attrs:
            return dict.__getattribute__(self, name)

        return self[name]

    def __setattr__(self, name, value):
        if name in Vertex._all_attrs:
            return dict.__setattr__(self, name, value)

        self[name] = value

    def __setitem__(self, name:str, value):
        dict.__setitem__(self, name, value)

        for id_parent_array, index_set in self._array_index_map.items():
            parent_array = di(id_parent_array)
            if name not in parent_array._attr_list_map:
                parent_array._attr_list_map[name] = AttrList(dtype=type(value))

            for index in index_set:
                parent_array._attr_list_map[name][index] = value

    def __getitem__(self, name:str):
        for id_parent_array, index_set in self._array_index_map.items():
            parent_array = di(id_parent_array)
            if name not in parent_array._attr_list_map:
                value = dict.__getitem__(self, name)
                parent_array._attr_list_map[name] = AttrList(dtype=type(value))
                for index in index_set:
                    parent_array._attr_list_map[name][index] = value

            for index in index_set:
                value = parent_array._attr_list_map[name][index]
                dict.__setitem__(self, name, value)
                return value
            
        return dict.__getitem__(self, name)

dtype_uint64 = np.array([], dtype=np.uint64).dtype
class Vertices:

    element_type = Vertex

    @checktype
    def __init__(self, _list:list=None, draw_type:GLInfo.draw_types=GL.GL_STATIC_DRAW, **kwargs):
        self._attr_list_map = {}
        self._vao_map = {}
        self._index_vertex_map = {}
        self._draw_type = draw_type
        self._tested_front_transparent = False
        self._tested_back_transparent = False
        self._front_has_transparent = False
        self._front_has_opaque = True
        self._back_has_transparent = False
        self._back_has_opaque = True
        
        if _list is not None:
            for value in _list:
                self.append(value)
        else:
            self._attr_list_map = kwargs

    def hasattr(self, attr_name:str):
        return (attr_name in self._attr_list_map)
    
    def reset(self, **kwargs):
        self._attr_list_map = kwargs
        self._vao_map = {}
        self._index_vertex_map = {}
        self._tested_front_transparent = False
        self._tested_back_transparent = False
        self._front_has_transparent = False
        self._front_has_opaque = True
        self._back_has_transparent = False
        self._back_has_opaque = True

    @property
    def draw_type(self):
        return self._draw_type

    @draw_type.setter
    @checktype
    def draw_type(self, value:GLInfo.draw_types):
        self._draw_type = value

    def _first_apply(self, program, instances)->bool:
        current_context = GLConfig.buffered_current_context
        key = (current_context, program, instances)
        if key in self._vao_map or not self:
            return False

        vao = VAO()
        self._vao_map[key] = vao
        return self._apply_increment(instances)

    def _update_VAOs(self, key, attr_list, divisor=None):
        for (context, program, insts), vao in self._vao_map.items():
            if key not in program._attributes_info:
                continue

            location = program._attributes_info[key]["location"]
            if location in vao and not attr_list.is_new_vbo:
                continue

            feed_type = attr_list.dtype
            need_type = program._attributes_info[key]["python_type"]
            if feed_type != need_type:
                if feed_type in GLInfo.primary_types and need_type in GLInfo.primary_types:
                    attr_list.dtype = need_type
                    feed_type = need_type
                elif feed_type == int and need_type == glm.uvec2:
                    attr_list.dtype = np.uint64
                    feed_type = glm.uvec2
                elif feed_type in (np.uint64, dtype_uint64) and need_type == glm.uvec2:
                    feed_type = glm.uvec2
                else:
                    error_message = f"vertex attribute '{key}' need type {need_type}, {feed_type} value were given"
                    raise TypeError(error_message)
            
            vao[location].interp(attr_list._vbo, feed_type, attr_list.stride, 0)
            if divisor is not None:
                vao[location].divisor = divisor

            attr_list.is_new_vbo = False

    def _apply_increment(self, instances)->bool:
        for key, attr_list in self._attr_list_map.items():
            attr_list._apply()
            self._update_VAOs(key, attr_list)

        if instances is None:
            return True
        
        for key, attr_list in instances._attr_list_map.items():
            attr_list._apply()
            self._update_VAOs(key, attr_list, instances.divisor)

        return True

    def _apply(self, program, instances):
        success = False
        current_context = GLConfig.buffered_current_context
        key = (current_context, program, instances)
        if key not in self._vao_map:
            success = self._first_apply(program, instances)
        else:
            success = self._apply_increment(instances)

        if success:
            self._vao_map[key].bind()

    def _process_slice(self, index):
        len_self = len(self)
        start = index.start if index.start is not None else 0
        stop = index.stop if index.stop is not None else len_self
        step = index.step if index.step is not None else 1

        if len_self > 0:
            while start < 0:
                start += len_self

            while stop < 0:
                stop += len_self

        if start < 0: start = 0
        if stop > len_self: stop = len_self

        return start, stop, step

    def __len__(self):        
        for attr_list in self._attr_list_map.values():
            return len(attr_list)
        
        return 0
    
    def __bool__(self):
        for attr_list in self._attr_list_map.values():
            return bool(attr_list)
        
        return False

    def append(self, vertex):
        len_self = len(self)
        for key in set.union(set(vertex.keys()), set(self._attr_list_map.keys())):
            if key not in self._attr_list_map:
                self._attr_list_map[key] = AttrList(dtype=type(vertex[key]))
            if key not in vertex:
                vertex[key] = self._attr_list_map[key].dtype()

            self._attr_list_map[key][len_self] = vertex[key]

        vertex._add_array_index(self, len_self)
        self._index_vertex_map[len_self] = vertex

    def __contains__(self, value:(Vertex,str)):
        if isinstance(value, str):
            return (value in self._attr_list_map)
        else:
            return (self.index(value) != -1)

    def __iter__(self):
        return SameTypeList.iterator(self)
    
    def const_items(self):
        return SameTypeList.const_iterator(self)

    def const_get(self, index:(int,slice)):
        result = None
        if isinstance(index, slice):
            result = []
            start, stop, step = self._process_slice(index)
            for sub_index in range(start, stop, step):
                result.append(self.const_get(sub_index))
        else:
            result = self.__class__.element_type()
            for key, attr_list in self._attr_list_map.items():
                result[key] = attr_list[index]

        return result

    def __getitem__(self, index:(int,slice,str)):
        if isinstance(index, str):
            return self._attr_list_map[index]

        result = None
        if isinstance(index, slice):
            result = []
            start, stop, step = self._process_slice(index)
            for sub_index in range(start, stop, step):
                result.append(self[sub_index])
        else:
            if index in self._index_vertex_map:
                result = self._index_vertex_map[index]
            else:
                result = self.__class__.element_type()
                result._add_array_index(self, index)
                for key, attr_list in self._attr_list_map.items():
                    if index >= len(attr_list):
                        result[key] = attr_list.dtype()
                    else:
                        result[key] = attr_list[index]
                self._index_vertex_map[index] = result

        return result

    def __setitem__(self, index:(int,slice), value:Vertex):
        for key in set.union(set(value.keys()), set(self._attr_list_map.keys())):
            if key not in self._attr_list_map:
                self._attr_list_map[key] = AttrList(dtype=type(value[key]))
            if key not in value:
                if len(self._attr_list_map[key]) == 0:
                    value[key] = self._attr_list_map[key].dtype()
                else:
                    value[key] = copy.deepcopy(self._attr_list_map[key].const_get(-1))

            self._attr_list_map[key][index] = value[key]

        if isinstance(index, int):
            value._add_array_index(self, index)
            self._index_vertex_map[index] = value
        elif isinstance(index, slice):
            start, stop, step = self._process_slice(index)
            for sub_index in range(start, stop, step):
                value._add_array_index(self, sub_index)
                self._index_vertex_map[sub_index] = value
        else:
            raise TypeError(index)

    def __delitem(self, index:int):
        if index in self._index_vertex_map:
            vertex = self._index_vertex_map[index]
            vertex._array_index_map[self].remove(index)
            if not vertex._array_index_map[self]:
                del vertex._array_index_map[self]

            del self._index_vertex_map[index]

        should_update_index = []
        for sub_index, vertex in self._index_vertex_map:
            if sub_index > index:
                should_update_index.append(sub_index)
                vertex._array_index_map[self].remove(sub_index)
                vertex._array_index_map[self].add(sub_index-1)

        for sub_index in should_update_index:
            vertex = self._index_vertex_map[sub_index]
            del self._index_vertex_map[sub_index]
            self._index_vertex_map[sub_index-1] = vertex

    def __delitem__(self, index:(str,int,slice)):
        if isinstance(index, str):
            del self._attr_list_map[index]
            return

        for key in self._attr_list_map:
            del self._attr_list_map[key][index]

        if isinstance(index, int):
            self.__delitem(index)
        else:
            start, stop, step = self._process_slice(index)
            for sub_index in range(start, stop, step):
                self.__delitem(sub_index)

    def clear(self):
        for key in self._attr_list_map:
            self._attr_list_map[key].clear()

        for vertex in self._index_vertex_map:
            del vertex._array_index_map[self]

        self._index_vertex_map.clear()

    def pop(self, index:int)->Vertex:
        vertex = self[index]
        del self[index]

        return vertex
    
    def index(self, vertex:Vertex)->int:
        for key in vertex:
            if key not in self._attr_list_map:
                return -1

        for i in range(len(self)):
            is_equal = True
            for key in vertex:
                if self._attr_list_map[key].const_get(i) != vertex[key]:
                    is_equal = False
                    break
            if is_equal:
                return i
            
        return -1
    
    def remove(self, vertex:Vertex):
        i = self.index(vertex)
        if i == -1:
            raise ValueError(vertex)
        
        del self[i]
    
    def extend(self, _list):
        for vertex in _list:
            self.append(vertex)

    def insert(self, index:int, vertex:Vertex)->None:
        for key in set.union(set(vertex.keys()), set(self._attr_list_map.keys())):
            if key not in vertex:
                vertex[key] = self._attr_list_map[key].dtype()

            vertex_attr = vertex[key]
            if key not in self._attr_list_map:
                self._attr_list_map[key] = AttrList(dtype=type(vertex_attr))
                delta_len = len(self) - len(self._attr_list_map[key])
                self._attr_list_map[key].extend([vertex_attr] * delta_len)
            else:
                self._attr_list_map[key].insert(index, vertex_attr)

        should_update_index = []
        for sub_index, sub_vertex in self._index_vertex_map.items():
            if sub_index >= index:
                should_update_index.append(sub_index)
                sub_vertex._array_index_map[self].remove(sub_index)
                sub_vertex._array_index_map[self].add(sub_index+1)

        for sub_index in should_update_index:
            sub_vertex = self._index_vertex_map[sub_index]
            del self._index_vertex_map[sub_index]
            self._index_vertex_map[sub_index+1] = sub_vertex

        vertex._add_array_index(self, index)
        self._index_vertex_map[index] = self

    def update(self, _list):
        len_list = len(_list)
        len_self = len(self)
        if len_list < len_self:
            del self[len_list:]
        
        for i in range(len_self):
            self[i] = _list[i]

        if len_list > len_self:
            self.extend(_list[len_self:])

    def _test_front_transparent(self):
        if "color" in self._attr_list_map:
            if not self._tested_front_transparent or self._attr_list_map["color"]._should_retest:
                front_alpha = self._attr_list_map["color"].ndarray[:,3]
                self._front_has_transparent = np.any(front_alpha < 1-1E-6)
                self._front_has_opaque = np.any(front_alpha >= 1-1E-6)
                self._attr_list_map["color"]._should_retest = False
        else:
            self._front_has_transparent = False
            self._front_has_opaque = True

        self._tested_front_transparent = True

    def _test_back_transparent(self):
        if "back_color" in self._attr_list_map:
            if not self._tested_back_transparent or self._attr_list_map["back_color"]._should_retest:
                back_alpha = self._attr_list_map["back_color"].ndarray[:,3]
                self._back_has_transparent = np.any(back_alpha < 1-1E-6)
                self._back_has_opaque = np.any(back_alpha >= 1-1E-6)
                self._attr_list_map["back_color"]._should_retest = False
        else:
            self._back_has_transparent = False
            self._back_has_opaque = True

        self._tested_back_transparent = True

    @property
    def front_has_transparent(self):
        self._test_front_transparent()
        return self._front_has_transparent
    
    @property
    def front_has_opaque(self):
        self._test_front_transparent()
        return self._front_has_opaque

    @property
    def back_has_transparent(self):
        self._test_back_transparent()
        return self._back_has_transparent
    
    @property
    def back_has_opaque(self):
        self._test_back_transparent()
        return self._back_has_opaque


import numpy as np
from OpenGL import GL

from .WeakSet import WeakSet
from .GLConfig import GLConfig

class _MetaGLObject(type):

    _all_instances = {}

    @property
    def active_id(cls):
        try:
            return GL.glGetIntegerv(cls._basic_info["binding_type"])
        except:
            return 0
        
    @property
    def all_instances(cls):
        if cls not in _MetaGLObject._all_instances:
            _MetaGLObject._all_instances[cls] = WeakSet()

        return _MetaGLObject._all_instances[cls]

class GLObject(metaclass=_MetaGLObject):

    def __init__(self, context_shared=True):
        self._id = 0
        self._context_shared = context_shared
        self._context = 0
        if self.__class__ not in _MetaGLObject._all_instances:
            _MetaGLObject._all_instances[self.__class__] = WeakSet()

        _MetaGLObject._all_instances[self.__class__].add(self)

    def __hash__(self)->int:
        return id(self)
    
    def __eq__(self, _value)->bool:
        return (id(self) == id(_value))

    @property
    def context_shared(self):
        return self._context_shared
    
    @property
    def context(self):
        return self._context

    @property
    def id(self):
        return self._id
    
    def delete(self):
        if self._id == 0:
            return

        try:
            self.unbind()
        except:
            pass

        try:
            del_func = self.__class__._basic_info["del_func"]
            need_number = self.__class__._basic_info["need_number"]
            if need_number:
                del_func(1, np.array([self._id]))
            else:
                del_func(self._id)
        except:
            pass

        self._id = 0

    def __del__(self):
        if self.__class__ in _MetaGLObject._all_instances:
            _MetaGLObject._all_instances[self.__class__].remove(self)

        if self._id == 0:
            return

        is_bound = False
        try:
            is_bound = self.is_bound
        except:
            pass

        if is_bound:
            try:
                target_type = self.__class__._basic_info["target_type"]
                bind_func = self.__class__._basic_info["bind_func"]
                if target_type is None:
                    bind_func(0)
                else:
                    bind_func(target_type, 0)
            except:
                pass

        try:
            del_func = self.__class__._basic_info["del_func"]
            need_number = self.__class__._basic_info["need_number"]
            if need_number:
                del_func(1, np.array([self._id]))
            else:
                del_func(self._id)
        except:
            pass

        self._id = 0

    def bind(self):
        current_context = GLConfig.buffered_current_context
        if self._context != 0 and current_context != self._context:
            raise RuntimeError(f"try to bind context({self._context})'s {self.__class__.__name__}({self.id}) in context({current_context})")

        if self._id == 0:
            gen_func = self.__class__._basic_info["gen_func"]
            need_number = self.__class__._basic_info["need_number"]
            if need_number:
                self._id = gen_func(1)
            else:
                self._id = gen_func()
            self._id = int(self._id)

            if self._id == 0:
                raise MemoryError("failed to create " + self.__class__.__name__)

        # if self.is_bound:
        #     return

        target_type = self.__class__._basic_info["target_type"]
        bind_func = self.__class__._basic_info["bind_func"]
        if target_type is None:
            bind_func(self._id)
        else:
            bind_func(target_type, self._id)

        if not self._context_shared:
            self._context = current_context

    def unbind(self):
        if not self.is_bound:
            return False

        target_type = self.__class__._basic_info["target_type"]
        bind_func = self.__class__._basic_info["bind_func"]
        try:
            if target_type is None:
                bind_func(0)
            else:
                bind_func(target_type, 0)
            return True
        except:
            return False

    @property
    def is_bound(self):
        return (self._id != 0 and self.__class__.active_id == self._id)
    


from .ShaderEffect import ShaderEffect
from .. import lut
from glass import sampler2D
from glass.utils import extname

import numpy as np
import os
from OpenGL import GL

class LUTEffect(ShaderEffect):

    def __init__(self, LUT:(str, np.ndarray, sampler2D), contribute:float=1.0):
        self_folder = os.path.dirname(os.path.abspath(__file__))
        ShaderEffect.__init__(self, self_folder + "/../glsl/PostProcessEffects/lut.glsl")

        if isinstance(LUT, str) and extname(LUT) == "cube":
            LUT = lut.cube_to_LUT(LUT)
        
        if not isinstance(LUT, sampler2D):
            LUT = sampler2D(LUT)

        LUT.wrap = GL.GL_CLAMP_TO_EDGE
        self["LUT"] = LUT
        self["contribute"] = contribute
        self.__LUT = LUT
        self.__contribute = contribute

    @property
    def LUT(self):
        return self.__LUT
    
    @LUT.setter
    def LUT(self, LUT:(str, np.ndarray, sampler2D)):
        if isinstance(LUT, str) and extname(LUT) == "cube":
            LUT = lut.cube_to_LUT(LUT)
        
        if not isinstance(LUT, sampler2D):
            LUT = sampler2D(LUT)

        LUT.wrap = GL.GL_CLAMP_TO_EDGE
        self.__LUT = LUT
        self["LUT"] = LUT

    @property
    def contribute(self):
        return self.__contribute
    
    @contribute.setter
    def contribute(self, contribute:float):
        self.__contribute = contribute
        self["contribute"] = contribute



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import numpy as np

class Floor(Mesh):

    __default_image = None

    @checktype
    def __init__(self, color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 length:float=1000, repeat_per_meter:float=1, name:str=""):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=True)
        self.__length = length
        self.__repeat_per_meter = repeat_per_meter

        if Floor.__default_image is None:
            Floor.__default_image = np.zeros((100, 100, 3), dtype=np.uint8)
            Floor.__default_image[:50, :50, :] = 140
            Floor.__default_image[50:, 50:, :] = 140
            Floor.__default_image[:50, 50:, :] = 127
            Floor.__default_image[50:, :50, :] = 127

        self.material.diffuse_map = Floor.__default_image
        self.material.cast_shadows = False
        self.start_building()

    def build(self):
        self.is_closed = False
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices
        length = self.__length
        repeat_per_meter = self.__repeat_per_meter
        vertices = self.vertices
        indices = self.indices

        vertices[0] = Vertex(
            position=glm.vec3(-length/2, -length/2, -0.001),
            tangent=glm.vec3(1/repeat_per_meter, 0, 0),
            bitangent=glm.vec3(0, 1/repeat_per_meter, 0),
            normal=glm.vec3(0, 0, 1),
            tex_coord=glm.vec3(0))
        vertices[1] = Vertex(
            position=glm.vec3( length/2, -length/2, -0.001),
            tangent=glm.vec3(1/repeat_per_meter, 0, 0),
            bitangent=glm.vec3(0, 1/repeat_per_meter, 0),
            normal=glm.vec3(0, 0, 1),
            tex_coord=repeat_per_meter*glm.vec3(length/2, 0, 0))
        vertices[2] = Vertex(
            position=glm.vec3( length/2,  length/2, -0.001),
            tangent=glm.vec3(1/repeat_per_meter, 0, 0),
            bitangent=glm.vec3(0, 1/repeat_per_meter, 0),
            normal=glm.vec3(0, 0, 1),
            tex_coord=repeat_per_meter*glm.vec3(length/2, length/2, 0))
        vertices[3] = Vertex(
            position=glm.vec3(-length/2,  length/2, -0.001),
            tangent=glm.vec3(1/repeat_per_meter, 0, 0),
            bitangent=glm.vec3(0, 1/repeat_per_meter, 0),
            normal=glm.vec3(0, 0, 1),
            tex_coord=repeat_per_meter*glm.vec3(0, length/2, 0))

        indices[0] = glm.uvec3(0, 1, 2)
        indices[1] = glm.uvec3(0, 2, 3)

        del vertices[4:]
        del indices[2:]

    @property
    def length(self):
        return self.__length
    
    @length.setter
    @Mesh.param_setter
    def length(self, length:glm.vec3):
        self.__length = length

    @property
    def repeat_per_meter(self):
        return self.__repeat_per_meter
    
    @repeat_per_meter.setter
    @Mesh.param_setter
    def repeat_per_meter(self, repeat_per_meter:int):
        self.__repeat_per_meter = repeat_per_meter



from .PointLight import PointLight, FlatPointLight
from glass.utils import checktype
from glass.DictList import DictList
from glass.ShaderStorageBlock import ShaderStorageBlock

import glm
import math

class SpotLight(PointLight):

    def __init__(self, name:str=""):
        PointLight.__init__(self, name)
        self._span_angle = 30
        self._softness = 10
        self.__update_aggregate_coeff()

    def __update_aggregate_coeff(self):
        half_span_angle_rad = self._span_angle/180*math.pi/2
        self._aggregate_coeff = 0.2/(1.1-math.cos(half_span_angle_rad))
        for flat in self._flats:
            flat.aggregate_coeff = self._aggregate_coeff

        self._update_scene_lights()

    @property
    def span_angle(self):
        return self._span_angle

    @span_angle.setter
    @checktype
    def span_angle(self, span_angle:float):
        if self._span_angle == span_angle:
            return

        self._span_angle = span_angle
        for flat in self._flats:
            flat.span_angle = span_angle

        self.__update_aggregate_coeff()

    @property
    def aggregate_coeff(self):
        return self._aggregate_coeff

    @property
    def softness(self):
        return self._softness
    
    @softness.setter
    @checktype
    def softness(self, softness:float):
        if self._softness == softness:
            return

        self._softness = softness
        for flat in self._flats:
            flat.softness = softness

        self._update_scene_lights()

class FlatSpotLight(FlatPointLight):

    def __init__(self, spot_light:SpotLight):
        self.direction = glm.vec3(0, 1, 0)
        FlatPointLight.__init__(self, spot_light)
        
    def update(self, spot_light:SpotLight):
        self.half_span_angle_rad = spot_light._span_angle/180*math.pi/2
        self.half_softness_rad = spot_light._softness/180*math.pi/2
        self.aggregate_coeff = spot_light.aggregate_coeff
        FlatPointLight.update(self, spot_light)

class SpotLights(ShaderStorageBlock.HostClass):

    def __init__(self):
        ShaderStorageBlock.HostClass.__init__(self)
        self.spot_lights = DictList()

    def __getitem__(self, path_str:str):
        return self.spot_lights[path_str]
    
    @ShaderStorageBlock.HostClass.not_const
    def __setitem__(self, path_str:str, spot_light:FlatSpotLight):
        self.spot_lights[path_str] = spot_light

    @ShaderStorageBlock.HostClass.not_const
    def __delitem__(self, path_str:str):
        del self.spot_lights[path_str]

    def __contains__(self, path_str:str):
        return (path_str in self.spot_lights)
    
    def __len__(self):
        return len(self.spot_lights)
    
    def __iter__(self):
        return iter(self.spot_lights)

    def keys(self):
        return self.spot_lights.keys()

    @property
    def n_spot_lights(self):
        return len(self.spot_lights)


from glass import GLConfig

class RenderHints:

    inherit = "inherit"
    __all_attrs = \
    {
        '__class__', '__delattr__', '__dict__', '__dir__', 
        '__doc__', '__enter__', '__eq__', '__exit__', '__format__', '__ge__', 
        '__getattribute__', '__getstate__', '__gt__', '__hash__', '__init__', 
        '__init_subclass__', '__le__', '__lt__', '__module__', '__ne__', 
        '__new__', '__reduce__', '__reduce_ex__', '__repr__', '__setattr__', 
        '__sizeof__', '__str__', '__subclasshook__', '__weakref__', '__getattr__',
        '_temp_env', 
    }

    def __init__(self):
        self._temp_env = GLConfig.LocalConfig()

    def __enter__(self):
        self._temp_env.__enter__()

    def __exit__(self, *exc_details):
        self._temp_env.__exit__(*exc_details)

    def __getattr__(self, name):
        if name in RenderHints.__all_attrs:
            return super().__getattr__(name)
        
        if name in self._temp_env:
            return self._temp_env[name]
        else:
            return RenderHints.inherit
    
    def __setattr__(self, name, value):
        if name in RenderHints.__all_attrs:
            return super().__setattr__(name, value)
        
        if value == RenderHints.inherit:
            del self._temp_env[name]
        else:
            self._temp_env[name] = value


from .SceneNode import SceneNode
from .Material import Material
from .algorithm import generate_auto_TBN, generate_smooth_TBN
from .callback_vec import callback_vec4

from glass import ShaderProgram, Instances, Vertices, Indices, GLInfo, RenderHints
from glass.utils import checktype, md5s
from glass.AttrList import AttrList

import glm
from functools import wraps
from OpenGL import GL
import inspect
import types
from enum import Enum
import copy
import numpy as np

class Mesh(SceneNode):

    __geometry_map = {}
    __builder_map = {}
    __mesh_vars = None

    class SurfType(Enum):
        Auto = 0
        Smooth = 1
        Flat = 2

    @checktype
    def __init__(self, primitive_type:GLInfo.primitive_types=GL.GL_TRIANGLES,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 name:str="", block:bool=True, shared:bool=True,
                 auto_build:bool=True, surf_type:SurfType=None):
        SceneNode.__init__(self, name)

        draw_type = GL.GL_DYNAMIC_DRAW if not block else GL.GL_STATIC_DRAW
        self._vertices = Vertices(draw_type=draw_type)
        self._indices = Indices(draw_type=draw_type)

        if isinstance(color, glm.vec3):
            color = glm.vec4(color, 1)

        if isinstance(back_color, glm.vec3):
            back_color = glm.vec4(back_color, 1)

        self._color = callback_vec4(color.r, color.g, color.b, color.a, self._color_change_callback)
        self._back_color_user_set = (back_color is not None)
        if self._back_color_user_set:
            self._back_color = callback_vec4(back_color.r, back_color.g, back_color.b, back_color.a, self._back_color_change_callback)
        else:
            self._back_color = callback_vec4(color.r, color.g, color.b, color.a, self._back_color_change_callback)
        
        self._should_add_color = True
        self._should_callback = True
        
        self._material = Material()
        self._material._opacity = 0
        self._material._parent_meshes.add(self)
        self._back_material = self._material
        self._back_material_user_set = False
        self._render_hints = RenderHints()
        
        self._propagation_props["explode_distance"] = 0

        self.__block = block
        self.__shared = shared
        self.__auto_build = auto_build
        self.__surf_type = surf_type
        self.__primitive = primitive_type
        self.__geometry_info = None
        self.__self_calculated_normal = False
        self.__has_transparent = False
        self.__has_opaque = True

    def __hash__(self):
        return id(self)

    def _set_transform_dirty(self, scenes):
        self._transform_dirty.update(scenes)
        return True

    @property
    def self_calculated_normal(self):
        return self.__self_calculated_normal
    
    @self_calculated_normal.setter
    def self_calculated_normal(self, flag:bool):
        self.__self_calculated_normal = flag

    @property
    def is_sphere(self):
        return (self.__class__.__name__ == "Sphere" and \
                self.scale.x == self.scale.y == self.scale.z and \
                self.span_lat >= 180 and self.span_lon >= 360) or \
               (self.__class__.__name__ == "Icosphere" and \
                self.scale.x == self.scale.y == self.scale.z)

    @property
    def has_transparent(self):
        return self.__has_transparent
    
    @property
    def has_opaque(self):
        return self.__has_opaque
    
    @property
    def is_filled(self):
        return (self.primitive_type in GLInfo.triangle_types)

    def build(self):
        pass

    @checktype
    def explode(self, distance:float):
        self.set_propagation_prop("explode_distance", distance)

    @property
    def explode_distance(self):
        return self.propagation_prop("explode_distance")
    
    @explode_distance.setter
    @checktype
    def explode_distance(self, distance:float):
        self.set_propagation_prop("explode_distance", distance)

    @staticmethod
    def _update_mesh_callback(child):
        if not isinstance(child, Mesh):
            return
        
        for scene in child.scenes:
            scene._update_mesh(child)

    @property
    def should_add_color(self):
        return self._should_add_color
    
    @should_add_color.setter
    def should_add_color(self, flag:bool):
        self._should_add_color = flag

    def __set_color(self):
        if not self._should_add_color:
            self._test_transparent()
            return

        color_array = np.dot(np.ones((len(self.vertices), 1), dtype=np.float32),
                             np.array([[self._color.r, self._color.g, self._color.b, self._color.a]], dtype=np.float32))
        if "color" not in self.vertices._attr_list_map:
            self.vertices._attr_list_map["color"] = AttrList(color_array, dtype=glm.vec4)
        else:
            self.vertices._attr_list_map["color"].ndarray = color_array

        back_color_array = np.dot(np.ones((len(self.vertices), 1), dtype=np.float32),
                                  np.array([[self._back_color.r, self._back_color.g, self._back_color.b, self._back_color.a]], dtype=np.float32))
        if "back_color" not in self.vertices._attr_list_map:
            self.vertices._attr_list_map["back_color"] = AttrList(back_color_array, dtype=glm.vec4)
        else:
            self.vertices._attr_list_map["back_color"].ndarray = back_color_array

        self._test_transparent()

    def _color_change_callback(self):
        if not self._should_callback:
            return
        
        self._should_callback = False

        color_array = np.dot(np.ones((len(self.vertices), 1), dtype=np.float32),
                             np.array([[self._color.r, self._color.g, self._color.b, self._color.a]], dtype=np.float32))
        if "color" not in self.vertices._attr_list_map:
            self.vertices._attr_list_map["color"] = AttrList(color_array, dtype=glm.vec4)
        else:
            self.vertices._attr_list_map["color"].ndarray = color_array

        if not self._back_color_user_set:
            if "back_color" not in self.vertices._attr_list_map:
                self.vertices._attr_list_map["back_color"] = AttrList(color_array, dtype=glm.vec4)
            else:
                self.vertices._attr_list_map["back_color"].ndarray = color_array

        self._test_transparent()

        self._should_callback = True

    @property
    def color(self)->callback_vec4:
        old_should_callback = self._should_callback
        self._should_callback = False
        if "color" not in self.vertices._attr_list_map:
            self._color.r = 0
            self._color.g = 0
            self._color.b = 0
            self._color.a = 0
        else:
            color_array = self.vertices._attr_list_map["color"].ndarray
            if len(color_array.shape) != 2 or color_array.shape[1] != 4:
                color_array = color_array.reshape(-1, 4)

            if color_array.size > 0:
                if np.all(color_array[:, 0] == color_array[0, 0]):
                    self._color.r = color_array[0, 0]
                else:
                    self._color.r = 0

                if np.all(color_array[:, 1] == color_array[0, 1]):
                    self._color.g = color_array[0, 1]
                else:
                    self._color.g = 0

                if np.all(color_array[:, 2] == color_array[0, 2]):
                    self._color.b = color_array[0, 2]
                else:
                    self._color.b = 0

                if np.all(color_array[:, 3] == color_array[0, 3]):
                    self._color.a = color_array[0, 3]
                else:
                    self._color.a = 0

        self._should_callback = old_should_callback

        return self._color
    
    @color.setter
    def color(self, color:(glm.vec3,glm.vec4)):
        if isinstance(color, glm.vec3):
            color = glm.vec4(color, 1)

        color_array = np.dot(np.ones((len(self.vertices), 1), dtype=np.float32),
                             np.array([[color.r, color.g, color.b, color.a]], dtype=np.float32))
        if "color" not in self.vertices._attr_list_map:
            self.vertices._attr_list_map["color"] = AttrList(color_array, dtype=glm.vec4)
        else:
            self.vertices._attr_list_map["color"].ndarray = color_array

        if not self._back_color_user_set:
            if "back_color" not in self.vertices._attr_list_map:
                self.vertices._attr_list_map["back_color"] = AttrList(color_array, dtype=glm.vec4)
            else:
                self.vertices._attr_list_map["back_color"].ndarray = color_array

        self._test_transparent()

    def _back_color_change_callback(self):
        if not self._should_callback:
            return

        self._should_callback = False

        self._back_color_user_set = True

        color_array = np.dot(np.ones((len(self.vertices), 1), dtype=np.float32),
                             np.array([[self._back_color.r, self._back_color.g, self._back_color.b, self._back_color.a]], dtype=np.float32))
        if "back_color" not in self.vertices._attr_list_map:
            self.vertices._attr_list_map["back_color"] = AttrList(color_array, dtype=glm.vec4)
        else:
            self.vertices._attr_list_map["back_color"].ndarray = color_array

        self._test_transparent()

        self._should_callback = True

    @property
    def back_color(self)->callback_vec4:
        old_should_callback = self._should_callback
        self._should_callback = False
        if "back_color" not in self.vertices._attr_list_map:
            self._back_color.r = 0
            self._back_color.g = 0
            self._back_color.b = 0
            self._back_color.a = 0
        else:
            color_array = self.vertices._attr_list_map["back_color"].ndarray
            if len(color_array.shape) != 2 or color_array.shape[1] != 4:
                color_array = color_array.reshape(-1, 4)

            if color_array.size > 0:
                if np.all(color_array[:, 0] == color_array[0, 0]):
                    self._back_color.r = color_array[0, 0]
                else:
                    self._back_color.r = 0

                if np.all(color_array[:, 1] == color_array[0, 1]):
                    self._back_color.g = color_array[0, 1]
                else:
                    self._back_color.g = 0

                if np.all(color_array[:, 2] == color_array[0, 2]):
                    self._back_color.b = color_array[0, 2]
                else:
                    self._back_color.b = 0

                if np.all(color_array[:, 3] == color_array[0, 3]):
                    self._back_color.a = color_array[0, 3]
                else:
                    self._back_color.a = 0
            else:
                self._back_color.r = 0
                self._back_color.g = 0
                self._back_color.b = 0
                self._back_color.a = 0
        self._should_callback = old_should_callback

        return self._back_color
    
    @back_color.setter
    def back_color(self, color:(glm.vec3,glm.vec4)):
        if isinstance(color, glm.vec3):
            color = glm.vec4(color, 1)

        self._back_color_user_set = True
        color_array = np.dot(np.ones((len(self.vertices), 1), dtype=np.float32),
                             np.array([[color.r, color.g, color.b, color.a]], dtype=np.float32))
        if "back_color" not in self.vertices._attr_list_map:
            self.vertices._attr_list_map["back_color"] = AttrList(color_array, dtype=glm.vec4)
        else:
            self.vertices._attr_list_map["back_color"].ndarray = color_array

        self._test_transparent()

    @property
    def back_material(self):
        if self._back_material is self._material and not self._back_material_user_set:
            self._back_material = copy.deepcopy(self._material)
            self._back_material._parent_meshes.clear()
            self._back_material._parent_meshes.add(self)

        return self._back_material
    
    @back_material.setter
    @checktype
    def back_material(self, material:Material):
        if material is None:
            self._back_material = self._material
        elif self._back_material is not material:
            self._back_material._parent_meshes.remove(self)
            material._parent_meshes.add(self)
            self._back_material = material

        self._back_material_user_set = True

        self._test_transparent()

    @property
    def render_hints(self):
        return self._render_hints

    @staticmethod
    def param_setter(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self = args[0]
            value = args[1]

            equal = False
            try:
                lvalue = getattr(self, func.__name__)
                if type(lvalue) != type(value):
                    equal = False
                else:
                    equal = bool(getattr(self, func.__name__) == value)
            except:
                equal = False

            if equal:
                return

            safe_func = checktype(func)
            return_value = safe_func(*args, **kwargs)

            if self.auto_build:
                self.start_building()

            return return_value

        return wrapper

    @property
    def surf_type(self):
        return self.__surf_type
    
    @surf_type.setter
    @checktype
    def surf_type(self, surf_type:SurfType):
        self.__surf_type = surf_type

    @property
    def block(self):
        return self.__block
    
    @block.setter
    @checktype
    def block(self, flag:bool):
        self.__block = flag

    @property
    def shared(self):
        return self.__shared
    
    @shared.setter
    @checktype
    def shared(self, flag:bool):
        self.__shared = flag

    @property
    def auto_build(self):
        return self.__auto_build
    
    @auto_build.setter
    @checktype
    def auto_build(self, flag:bool):
        self.__auto_build = flag

    @property
    def bounding_box(self):
        return (self.x_min, self.x_max, self.y_min, self.y_max, self.z_min, self.z_max)

    @property
    def center(self):
        return glm.vec3(0.5*(self.x_min + self.x_max),
                        0.5*(self.y_min + self.y_max),
                        0.5*(self.z_min + self.z_max))

    @property
    def x_min(self):
        if self.__geometry_info is None:
            return 0
        
        return self.__geometry_info["x_min"]
    
    @x_min.setter
    @checktype
    def x_min(self, x_min:float):
        if self.__geometry_info is None:
            return
        
        self.__geometry_info["x_min"] = x_min
    
    @property
    def x_max(self):
        if self.__geometry_info is None:
            return 0
        
        return self.__geometry_info["x_max"]
    
    @x_max.setter
    @checktype
    def x_max(self, x_max:float):
        if self.__geometry_info is None:
            return
        
        self.__geometry_info["x_max"] = x_max
    
    @property
    def y_min(self):
        if self.__geometry_info is None:
            return 0
        
        return self.__geometry_info["y_min"]
    
    @y_min.setter
    @checktype
    def y_min(self, y_min:float):
        if self.__geometry_info is None:
            return
        
        self.__geometry_info["y_min"] = y_min
    
    @property
    def y_max(self):
        if self.__geometry_info is None:
            return 0
        
        return self.__geometry_info["y_max"]
    
    @y_max.setter
    @checktype
    def y_max(self, y_max:float):
        if self.__geometry_info is None:
            return
        
        self.__geometry_info["y_max"] = y_max
    
    @property
    def z_min(self):
        if self.__geometry_info is None:
            return 0
        
        return self.__geometry_info["z_min"]
    
    @z_min.setter
    @checktype
    def z_min(self, z_min:float):
        if self.__geometry_info is None:
            return
        
        self.__geometry_info["z_min"] = z_min

    @property
    def z_max(self):
        if self.__geometry_info is None:
            return 0
        
        return self.__geometry_info["z_max"]
    
    @z_max.setter
    @checktype
    def z_max(self, z_max:float):
        if self.__geometry_info is None:
            return
        
        self.__geometry_info["z_max"] = z_max

    def __copy_vertices(self, shared_vertices):
        if self._vertices is shared_vertices:
            return
        
        for key, value in shared_vertices._attr_list_map.items():
            if key not in ["color", "back_color"]:
                self._vertices._attr_list_map[key] = value

        self.__set_color()

    def start_building(self):
        if self.__class__.__name__ == "Mesh":
            return

        if self.__shared:
            instance_key = self.__instance_key
            if instance_key in Mesh.__geometry_map:
                geometry_info = Mesh.__geometry_map[instance_key]
                self.__copy_vertices(geometry_info["vertices"])
                self._indices = geometry_info["indices"]
                self.__geometry_info = geometry_info
                if not self.__block:
                    return

                builder = self.__geometry_info["builder"]
                if builder is None:
                    return
                
                try:
                    while True:
                        next(builder)
                except StopIteration:
                    del Mesh.__builder_map[builder]
                    self.__post_build(self.__geometry_info)
                    self.__geometry_info["builder"] = None

                return
            else:
                draw_type = GL.GL_DYNAMIC_DRAW if not self.__block else GL.GL_STATIC_DRAW
                geometry_info = \
                {
                    "vertices": Vertices(draw_type=draw_type),
                    "indices": Indices(draw_type=draw_type),
                    "builder": None,
                    "x_min": 0, "x_max": 0,
                    "y_min": 0, "y_max": 0,
                    "z_min": 0, "z_max": 0,
                }
                Mesh.__geometry_map[instance_key] = geometry_info
                self._vertices = geometry_info["vertices"]
                self._indices = geometry_info["indices"]
                self.__geometry_info = geometry_info
        elif self.__geometry_info is None:
            draw_type = GL.GL_DYNAMIC_DRAW if not self.__block else GL.GL_STATIC_DRAW
            self.__geometry_info = \
            {
                "vertices": Vertices(draw_type=draw_type),
                "indices": Indices(draw_type=draw_type),
                "builder": None,
                "x_min": 0, "x_max": 0,
                "y_min": 0, "y_max": 0,
                "z_min": 0, "z_max": 0,
            }
            self._vertices = self.__geometry_info["vertices"]
            self._indices = self.__geometry_info["indices"]

        if self.__block:
            if inspect.isgeneratorfunction(self.build):
                builder = self.build()
                try:
                    while True:
                        next(builder)
                except StopIteration:
                    self.__post_build(self.__geometry_info)
                    self.__geometry_info["builder"] = None
                
            else: # not generator
                self.build()
                self.__post_build(self.__geometry_info)
                self.__geometry_info["builder"] = None

        else: # not block
            if inspect.isgeneratorfunction(self.build):
                builder = self.build()
                self.__geometry_info["builder"] = builder
                Mesh.__builder_map[builder] = self.__geometry_info
                try:
                    next(builder)
                except StopIteration:
                    del Mesh.__builder_map[builder]
                    self.__post_build(self.__geometry_info)
                    self.__geometry_info["builder"] = None
            else:
                self.build()
                self.__post_build(self.__geometry_info)
                self.__geometry_info["builder"] = None

    @property
    def __instance_key(self):
        if Mesh.__mesh_vars is None:
            temp_mesh = Mesh()
            Mesh.__mesh_vars = set(temp_mesh.__dict__.keys())

        self_vars = set(self.__dict__.keys())
        new_vars = sorted(list(self_vars - Mesh.__mesh_vars))
        len_new_vars = len(new_vars)
        instance_key = self.__class__.__name__ + "("
        for i in range(len_new_vars):
            key = new_vars[i]
            value = self.__dict__[key]
            str_value = None
            if isinstance(value, (int,float,bool,complex,str,bytes,bytearray)):
                str_value = str(value)
            else:
                try:
                    str_value = f"md5({md5s(value)})"
                except:
                    str_value = f"id({id(value)})"

            instance_key += (key + "=" + str_value)
            if i < len_new_vars-1:
                instance_key += ", "
        instance_key += ")"
        return instance_key

    @property
    def is_building(self):
        return (self.__geometry_info is not None and self.__geometry_info["builder"] is not None)

    @property
    def is_generating(self):
        return (self.__geometry_info is not None and self.__geometry_info["builder"] is not None and isinstance(self.__geometry_info["builder"], types.GeneratorType))

    def generate(self):
        builder = self.__geometry_info["builder"]
        if builder is None:
            return
        
        try:
            next(builder)
        except StopIteration:
            del Mesh.__builder_map[builder]
            self.__post_build(self.__geometry_info)
            self.__geometry_info["builder"] = None
    
    def __eq__(self, other):
        return (id(self) == id(other))

    @property
    def vertices(self):
        return self._vertices

    @vertices.setter
    @checktype
    def vertices(self, vertices:(Vertices, list)):
        if self._vertices is vertices:
            return

        if isinstance(vertices, Vertices):
            self._vertices = vertices
        else:
            self._vertices = Vertices(vertices)

        self.__geometry_info["vertices"] = self._vertices

    @property
    def indices(self):
        return self._indices

    @indices.setter
    @checktype
    def indices(self, indices:(Indices, list)):
        if self._indices is indices:
            return
        
        if isinstance(indices, Indices):
            self._indices = indices
        else:
            self._indices = Indices(indices)

        self.__geometry_info["indices"] = self._indices

    @property
    def material(self):
        return self._material

    @material.setter
    @checktype
    def material(self, material:Material):
        if material is None:
            raise ValueError(material)

        if self._material is material:
            return

        self._material._parent_meshes.remove(self)
        material._parent_meshes.add(self)

        self._material = material
        
        if not material._opacity_user_set:
            material._opacity = 1

        if not self._back_color_user_set:
            self._back_material = material

        self._test_transparent()
    
    def __post_build(self, geometry_info):
        self.__set_color()
        self.__calculate_bounding_box(geometry_info)
        self.__generate_TBN(geometry_info)

    def __calculate_bounding_box(self, geometry_info):
        vertices = geometry_info["vertices"]
        if not vertices or "position" not in vertices:
            return
        
        if geometry_info["x_min"] == 0 and geometry_info["x_max"] == 0 and \
           geometry_info["y_min"] == 0 and geometry_info["y_max"] == 0 and \
           geometry_info["z_min"] == 0 and geometry_info["z_max"] == 0:
            geometry_info["x_min"] = vertices["position"].ndarray[:,0].min()
            geometry_info["x_max"] = vertices["position"].ndarray[:,0].max()
            geometry_info["y_min"] = vertices["position"].ndarray[:,1].min()
            geometry_info["y_max"] = vertices["position"].ndarray[:,1].max()
            geometry_info["z_min"] = vertices["position"].ndarray[:,2].min()
            geometry_info["z_max"] = vertices["position"].ndarray[:,2].max()

    @property
    def primitive_type(self):
        return self.__primitive

    @primitive_type.setter
    @checktype
    def primitive_type(self, primitive_type:GLInfo.primitive_types):
        self.__primitive = primitive_type

    def generate_temp_TBN(self, vertex0, vertex1, vertex2):
        v01 = vertex1.position - vertex0.position
        v02 = vertex2.position - vertex0.position

        st01 = vertex1.tex_coord - vertex0.tex_coord
        st02 = vertex2.tex_coord - vertex0.tex_coord

        det = st01.s * st02.t - st02.s * st01.t
        
        normal = glm.cross(v01, v02)
        len_normal = glm.length(normal)
        if len_normal > 1E-6:
            normal = normal / len_normal
        else:
            normal = glm.vec3(0)

        tangent = st02.t*v01 - st01.t*v02
        bitangent = st01.s*v02 - st02.s*v01
        if abs(det) > 1E-6:
            tangent /= det
            bitangent /= det
        else:
            tangent = glm.vec3(0)
            bitangent = glm.vec3(0)
        
        # vertex0
        vertex0.tangent = tangent
        vertex0.bitangent = bitangent
        
        # vertex1
        vertex1.tangent = tangent
        vertex1.bitangent = bitangent
        
        # vertex2
        vertex2.tangent = tangent
        vertex2.bitangent = bitangent

        if not self.self_calculated_normal:
            vertex0.normal = normal
            vertex1.normal = normal
            vertex2.normal = normal

    def __generate_TBN(self, geometry_info):        
        vertices = geometry_info["vertices"]
        indices = geometry_info["indices"]

        if not vertices or not indices or \
           'position' not in vertices or \
           'color' not in vertices:
            return

        if "tangent" not in vertices._attr_list_map:
            vertices._attr_list_map["tangent"] = AttrList(np.zeros_like(vertices["position"].ndarray), dtype=glm.vec3)

        if "bitangent" not in vertices._attr_list_map:
            vertices._attr_list_map["bitangent"] = AttrList(np.zeros_like(vertices["position"].ndarray), dtype=glm.vec3)

        if "normal" not in vertices._attr_list_map:
            vertices._attr_list_map["normal"] = AttrList(np.zeros_like(vertices["position"].ndarray), dtype=glm.vec3)

        if self.__surf_type == Mesh.SurfType.Auto:
            generate_auto_TBN(vertices, indices, not self.self_calculated_normal)
        elif self.__surf_type == Mesh.SurfType.Smooth:
            generate_smooth_TBN(vertices, indices, not self.self_calculated_normal)

    def draw(self, program:ShaderProgram, instances:Instances=None):
        if not self.visible:
            return

        with self.render_hints:
            if self.__primitive in GLInfo.triangle_types:
                program.draw_triangles(self._vertices, self._indices, instances, self.__primitive)
            elif self.__primitive in GLInfo.line_types:
                if self.__primitive in [GL.GL_LINE_STRIP, GL.GL_LINE_LOOP]:
                    program.draw_lines(self._vertices, None, instances, self.__primitive)
                else:
                    program.draw_lines(self._vertices, self._indices, instances, self.__primitive)
            elif self.__primitive == GL.GL_POINTS:
                program.draw_points(self._vertices, instances)
            elif self.__primitive == GL.GL_PATCHES:
                program.draw_patches(self._vertices, self._indices, instances)

    def _test_transparent(self):
        if not self._vertices:
            return

        front_has_transparent = self._vertices.front_has_transparent and self._material.has_transparent
        back_has_transparent = self._vertices.back_has_transparent and self._back_material.has_transparent
        self.__has_transparent = (front_has_transparent or back_has_transparent)

        front_has_opaque = self._vertices.front_has_opaque or self._material.has_opaque
        back_has_opaque = self._vertices.back_has_opaque or self._back_material.has_opaque
        self.__has_opaque = (front_has_opaque or back_has_opaque)


from glass.utils import pip_install
from .QtScreen import init_QtScreen
from ..Manipulators.Manipulator import Manipulator

try:
    import PyQt5
    from PyQt5.QtWidgets import QOpenGLWidget
except ModuleNotFoundError:
    pip_install("PyQt5")
    import PyQt5
    from PyQt5.QtWidgets import QOpenGLWidget

import glm

@init_QtScreen
class PyQt5Screen(QOpenGLWidget):
    mouse_pressed = PyQt5.QtCore.pyqtSignal(Manipulator.MouseButton, glm.vec2, glm.vec2)
    mouse_released = PyQt5.QtCore.pyqtSignal(Manipulator.MouseButton, glm.vec2, glm.vec2)
    mouse_double_clicked = PyQt5.QtCore.pyqtSignal(Manipulator.MouseButton, glm.vec2, glm.vec2)
    mouse_moved = PyQt5.QtCore.pyqtSignal(glm.vec2, glm.vec2)
    wheel_scrolled = PyQt5.QtCore.pyqtSignal(glm.vec2, glm.vec2, glm.vec2)
    key_pressed = PyQt5.QtCore.pyqtSignal(Manipulator.Key)
    key_released = PyQt5.QtCore.pyqtSignal(Manipulator.Key)
    key_repeated = PyQt5.QtCore.pyqtSignal(set)
    frame_started = PyQt5.QtCore.pyqtSignal()
    frame_ended = PyQt5.QtCore.pyqtSignal()



from .CommonRenderer import CommonRenderer
from ..Frame import Frame

from glass import \
    ShaderProgram, GLConfig, FBO, sampler2DMS, sampler2D, usampler2D, usampler2DMS

from OpenGL import GL
import glm
import os
        
class DeferredRenderer(CommonRenderer):

    def __init__(self):
        CommonRenderer.__init__(self)
        
    def startup(self):
        CommonRenderer.startup(self)
        screen = self.camera.screen
        if not screen._samples_set_by_user and not screen._is_gl_init:
            screen._set_samples(1)
        screen.FXAA = True

    @property
    def deferred_render_program(self):
        if "deferred_render" in self.programs:
            return self.programs["deferred_render"]
        
        program = ShaderProgram()
        program.compile(Frame.draw_frame_vs)
        program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/Pipelines/deferred_rendering/deferred_rendering.fs")
        program["PointLights"].bind(self.scene.point_lights)
        program["DirLights"].bind(self.scene.dir_lights)
        program["SpotLights"].bind(self.scene.spot_lights)
        
        self.programs["deferred_render"] = program

        return program

    @property
    def draw_to_gbuffer_program(self):
        if "draw_to_gbuffer" in self.programs:
            return self.programs["draw_to_gbuffer"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_rendering.vs")
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_rendering.gs")
        program.compile(self_folder + "/../glsl/Pipelines/deferred_rendering/draw_to_gbuffer.fs")
        program["PointLights"].bind(self.scene.point_lights)
        program["DirLights"].bind(self.scene.dir_lights)
        program["SpotLights"].bind(self.scene.spot_lights)
        self.programs["draw_to_gbuffer"] = program

        return program
    
    @property
    def draw_lines_to_gbuffer_program(self):
        if "draw_lines_to_gbuffer" in self.programs:
            return self.programs["draw_lines_to_gbuffer"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_draw_lines.vs")
        program.compile(self_folder + "/../glsl/Pipelines/deferred_rendering/draw_points_to_gbuffer.fs")
        program["PointLights"].bind(self.scene.point_lights)
        program["DirLights"].bind(self.scene.dir_lights)
        program["SpotLights"].bind(self.scene.spot_lights)
        self.programs["draw_lines_to_gbuffer"] = program

        return program

    @property
    def draw_points_to_gbuffer_program(self):
        if "draw_points_to_gbuffer" in self.programs:
            return self.programs["draw_points_to_gbuffer"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_draw_points.vs")
        program.compile(self_folder + "/../glsl/Pipelines/deferred_rendering/draw_points_to_gbuffer.fs")
        program["PointLights"].bind(self.scene.point_lights)
        program["DirLights"].bind(self.scene.dir_lights)
        program["SpotLights"].bind(self.scene.spot_lights)
        self.programs["draw_points_to_gbuffer"] = program

        return program
    
    @property
    def screen_fbo(self):
        fbo = None
        screen_size = GLConfig.screen_size
        if "screen" in self.fbos:
            fbo = self.fbos["screen"]
            fbo.resize(screen_size.x, screen_size.y)
        else:
            fbo = FBO(screen_size.x, screen_size.y)
            fbo.attach(0, sampler2D)
            self.fbos["screen"] = fbo

        return fbo

    @property
    def gbuffer(self):
        fbo = None
        screen_size = GLConfig.screen_size
        samples = self.camera.screen.samples
        if samples > 1:
            if "gbuffer_ms" in self.fbos:
                fbo = self.fbos["gbuffer_ms"]
                fbo.resize(screen_size.x, screen_size.y, samples)
            else:
                fbo = FBO(screen_size.x, screen_size.y, samples)
                fbo.attach(0, sampler2DMS, GL.GL_RGBA32F) # view_pos_and_alpha
                fbo.attach(1, sampler2DMS, GL.GL_RGBA32F) # view_normal_and_emission_r
                fbo.attach(2, sampler2DMS, GL.GL_RGBA32F) # ambient_and_emission_g
                fbo.attach(3, sampler2DMS, GL.GL_RGBA32F) # diffuse_or_base_color_and_emission_b
                fbo.attach(4, sampler2DMS, GL.GL_RGBA32F) # specular_or_preshading_and_shininess
                fbo.attach(5, sampler2DMS, GL.GL_RGBA32F) # reflection
                fbo.attach(6, sampler2DMS, GL.GL_RGBA32F) # env_center_and_mixed_value
                fbo.attach(7, usampler2DMS, GL.GL_RGBA32UI) # mixed_uint
                fbo.attach(GL.GL_DEPTH_ATTACHMENT, sampler2DMS)
                fbo.auto_clear = False
                self.fbos["gbuffer_ms"] = fbo
            return fbo
        else:
            if "gbuffer" in self.fbos:
                fbo = self.fbos["gbuffer"]
                fbo.resize(screen_size.x, screen_size.y)
            else:
                fbo = FBO(screen_size.x, screen_size.y)
                fbo.attach(0, sampler2D, GL.GL_RGBA32F) # view_pos_and_alpha
                fbo.attach(1, sampler2D, GL.GL_RGBA32F) # view_normal_and_emission_r
                fbo.attach(2, sampler2D, GL.GL_RGBA32F) # ambient_and_emission_g
                fbo.attach(3, sampler2D, GL.GL_RGBA32F) # diffuse_or_base_color_and_emission_b
                fbo.attach(4, sampler2D, GL.GL_RGBA32F) # specular_or_preshading_and_shininess
                fbo.attach(5, sampler2D, GL.GL_RGBA32F) # reflection
                fbo.attach(6, sampler2D, GL.GL_RGBA32F) # env_center_and_mixed_value
                fbo.attach(7, usampler2D, GL.GL_RGBA32UI) # mixed_uint
                fbo.attach(GL.GL_DEPTH_ATTACHMENT, sampler2D)
                fbo.auto_clear = False
                self.fbos["gbuffer"] = fbo
            return fbo

    def prepare_draw_to_gbuffer(self):
        self.draw_to_gbuffer_program["camera"] = self.camera

    def draw_to_gbuffer(self, mesh, instances):
        if mesh.material.need_env_map or mesh._back_material.need_env_map:
            self.gen_env_map(mesh, instances)

        self.draw_to_gbuffer_program["material"] = mesh.material
        self.draw_to_gbuffer_program["back_material"] = mesh._back_material
        self.draw_to_gbuffer_program["explode_distance"] = mesh.explode_distance
        self.draw_to_gbuffer_program["is_sphere"] = mesh.is_sphere
        self.draw_to_gbuffer_program["mesh_center"] = mesh.center
        mesh.draw(self.draw_to_gbuffer_program, instances)

    def prepare_draw_lines_to_gbuffer(self):
        self.draw_lines_to_gbuffer_program["camera"] = self.camera

    def draw_lines_to_gbuffer(self, mesh, instances):
        if mesh.material.need_env_map:
            self.gen_env_map(mesh, instances)

        self.draw_lines_to_gbuffer_program["material"] = mesh.material
        self.draw_lines_to_gbuffer_program["mesh_center"] = mesh.center
        mesh.draw(self.draw_lines_to_gbuffer_program, instances)

    def prepare_draw_points_to_gbuffer(self):
        self.draw_points_to_gbuffer_program["camera"] = self.camera

    def draw_points_to_gbuffer(self, mesh, instances):
        if mesh.material.need_env_map:
            self.gen_env_map(mesh, instances)

        self.draw_points_to_gbuffer_program["material"] = mesh.material
        self.draw_points_to_gbuffer_program["mesh_center"] = mesh.center
        mesh.draw(self.draw_points_to_gbuffer_program, instances)

    def draw_opaque(self):
        if not self._opaque_meshes and \
           not self._opaque_lines and \
           not self._opaque_points:
            return
        
        with GLConfig.LocalConfig(clear_color=glm.vec4(0)):
            with self.gbuffer:
                GLConfig.clear_buffers()

                if self._opaque_meshes:
                    self.prepare_draw_to_gbuffer()
                    for mesh, instances in self._opaque_meshes:
                        self.draw_to_gbuffer(mesh, instances)

                if self._opaque_lines:
                    self.prepare_draw_lines_to_gbuffer()
                    for mesh, instances in self._opaque_lines:
                        self.draw_lines_to_gbuffer(mesh, instances)

                if self._opaque_points:
                    self.prepare_draw_points_to_gbuffer()
                    for mesh, instances in self._opaque_points:
                        self.draw_points_to_gbuffer(mesh, instances)

        resolved = self.gbuffer.resolved
        view_pos_and_alpha_map = resolved.color_attachment(3)
        view_normal_and_emission_r_map = resolved.color_attachment(4)
        ambient_and_emission_g_map = resolved.color_attachment(2)
        diffuse_or_base_color_and_emission_b_map = resolved.color_attachment(0)
        specular_or_preshading_and_shininess_map = resolved.color_attachment(1)
        reflection_map = resolved.color_attachment(5)
        env_center_and_mixed_value_map = resolved.color_attachment(6)
        mixed_uint_map = resolved.color_attachment(7)

        self._view_pos_map = view_pos_and_alpha_map
        self._view_normal_map = view_normal_and_emission_r_map
        self._depth_map = resolved.depth_attachment

        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            GLConfig.clear_buffers()
            self.deferred_render_program["camera"] = self.camera
            self.deferred_render_program["view_pos_and_alpha_map"] = view_pos_and_alpha_map
            self.deferred_render_program["view_normal_and_emission_r_map"] = view_normal_and_emission_r_map
            self.deferred_render_program["ambient_and_emission_g_map"] = ambient_and_emission_g_map
            self.deferred_render_program["diffuse_or_base_color_and_emission_b_map"] = diffuse_or_base_color_and_emission_b_map
            self.deferred_render_program["specular_or_preshading_and_shininess_map"] = specular_or_preshading_and_shininess_map
            self.deferred_render_program["reflection_map"] = reflection_map
            self.deferred_render_program["env_center_and_mixed_value_map"] = env_center_and_mixed_value_map
            self.deferred_render_program["mixed_uint_map"] = mixed_uint_map
            self.deferred_render_program["background"] = self.scene.background
            self.deferred_render_program["fog"] = self.scene.fog
            self.deferred_render_program.draw_triangles(Frame.vertices, Frame.indices)

        self.gbuffer.draw_to_active(GL.GL_DEPTH_ATTACHMENT)

        # draw skybox
        if self.scene.skybox.is_completed:
            self.scene.skybox.draw(self.camera)

        # draw skydome
        elif self.scene.skydome.is_completed:
            self.scene.skydome.draw(self.camera)
        
        if self._transparent_meshes or self._transparent_lines or self._transparent_points:
            self.gbuffer.draw_to(self.OIT_fbo, GL.GL_DEPTH_ATTACHMENT)

    def render(self):
        self._should_update = False
        sampler2D._should_update = False
        self.classify_meshes()
        self.update_dir_lights_depth()
        self.update_point_lights_depth()
        self.update_spot_lights_depth()
        self.draw_opaque()
        self.draw_transparent()
        return (self._should_update or sampler2D._should_update)


from .PostProcessEffect import PostProcessEffect
from glass import ShaderProgram, FBO, sampler2D, GLInfo, samplerCube, sampler2DArray
from ..Frame import Frame

import os

class FXAAEffect(PostProcessEffect):

    __array_programs = {}
    __program = None
    __cube_program = None

    @staticmethod
    def program():
        if FXAAEffect.__program is None:
            FXAAEffect.__program = ShaderProgram()
            FXAAEffect.__program.compile(Frame.draw_frame_vs)
            FXAAEffect.__program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/FXAA.fs")
        return FXAAEffect.__program
    
    @staticmethod
    def cube_program():
        if FXAAEffect.__cube_program is None:
            FXAAEffect.__cube_program = ShaderProgram()
            FXAAEffect.__cube_program.compile(Frame.draw_frame_vs)
            FXAAEffect.__cube_program.compile(Frame.draw_frame_array_gs(6))
            FXAAEffect.__cube_program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/FXAA_cube_filter.fs")
        return FXAAEffect.__cube_program
    
    def __init__(self, internal_format:GLInfo.internal_formats=None):
        PostProcessEffect.__init__(self)
        
        self._internal_format = internal_format
        self._fbo = None
        self._array_fbo = None
        self._cube_fbo = None

    @property
    def fbo(self):
        if self._fbo is None:
            self._fbo = FBO()
            self._fbo.attach(0, sampler2D, self._internal_format)
        return self._fbo

    @property
    def array_fbo(self):
        if self._array_fbo is None:
            self._array_fbo = FBO()
            self._array_fbo.attach(0, sampler2DArray, self._internal_format)
        return self._array_fbo

    @property
    def cube_fbo(self):
        if self._cube_fbo is None:
            self._cube_fbo = FBO()
            self._cube_fbo.attach(0, samplerCube, self._internal_format)
        return self._cube_fbo

    def need_pos_info(self) -> bool:
        return False

    def apply(self, screen_image:(sampler2D,samplerCube,sampler2DArray))->(sampler2D,samplerCube,sampler2DArray):
        if isinstance(screen_image, sampler2D):
            self.fbo.resize(screen_image.width, screen_image.height)
            with self.fbo:
                FXAAEffect.program()["screen_image"] = screen_image
                FXAAEffect.program().draw_triangles(Frame.vertices, Frame.indices)
                
            return self.fbo.color_attachment(0)
        elif isinstance(screen_image, samplerCube):
            self.cube_fbo.resize(screen_image.width, screen_image.height)
            with self.cube_fbo:
                FXAAEffect.cube_program()["screen_image"] = screen_image
                FXAAEffect.cube_program().draw_triangles(Frame.vertices, Frame.indices)

            return self.cube_fbo.color_attachment(0)
        elif isinstance(screen_image, sampler2DArray):
            self.array_fbo.resize(screen_image.width, screen_image.height, layers=screen_image.layers)
            with self.array_fbo:
                program = FXAAEffect.array_program(screen_image.layers)
                program["screen_image"] = screen_image
                program.draw_triangles(Frame.vertices, Frame.indices)

            return self.array_fbo.color_attachment(0)
    
    def draw_to_active(self, screen_image: sampler2D) -> None:
        FXAAEffect.program()["screen_image"] = screen_image
        FXAAEffect.program().draw_triangles(Frame.vertices, Frame.indices)
    
    @staticmethod
    def array_program(layers):
        if layers in FXAAEffect.__array_programs:
            return FXAAEffect.__array_programs[layers]

        program = ShaderProgram()
        program.compile(Frame.draw_frame_vs)
        program.compile(Frame.draw_frame_array_gs(layers))
        program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/FXAA_array_filter.fs")

        FXAAEffect.__array_programs[layers] = program

        return program


from ..Mesh import Mesh
from ..ColorMap import ColorMap
from .Surf import Surf

from glass.utils import checktype

import numpy as np
import glm

class FSurf(Mesh):

    @checktype
    def __init__(self, func, x_range=[-3,3], y_range=[-3,3], z_range=[-3,3], dep_var="Z",
                 color_map:ColorMap=None, back_color_map:ColorMap=None,
                 color:(glm.vec3,glm.vec4)=None, back_color:(glm.vec3,glm.vec4)=None,
                 surf_type:Mesh.SurfType=Mesh.SurfType.Smooth,
                 name:str=""):
        Mesh.__init__(self, name=name, surf_type=surf_type)
        self._func = func
        self._x_range = x_range
        self._y_range = y_range
        self._z_range = z_range
        self._dep_var = dep_var.upper()
        
        Surf._set_colors(self, color, back_color, color_map, back_color_map)
        self.start_building()
            
    def build(self):
        func = self._func
        x_range = self._x_range
        y_range = self._y_range
        z_range = self._z_range
        dep_var = self._dep_var
        
        X, Y, Z = None, None, None
        if dep_var == "X":
            Y = np.linspace(y_range[0], y_range[1]) if len(y_range) == 2 else y_range
            Z = np.linspace(z_range[0], z_range[1]) if len(z_range) == 2 else z_range
            Y, Z = np.meshgrid(Y, Z)
            try:
                X = func(Y, Z)
            except:
                func = np.vectorize(func)
                X = func(Y, Z)
        elif dep_var == "Y":
            X = np.linspace(x_range[0], x_range[1]) if len(x_range) == 2 else x_range
            Z = np.linspace(z_range[0], z_range[1]) if len(z_range) == 2 else z_range
            X, Z = np.meshgrid(X, Z)
            try:
                Y = func(X, Z)
            except:
                func = np.vectorize(func)
                Y = func(X, Z)
        elif dep_var == "Z":
            X = np.linspace(x_range[0], x_range[1]) if len(x_range) == 2 else x_range
            Y = np.linspace(y_range[0], y_range[1]) if len(y_range) == 2 else y_range
            X, Y = np.meshgrid(X, Y)
            try:
                Z = func(X, Y)
            except:
                func = np.vectorize(func)
                Z = func(X, Y)
        C = eval(dep_var)

        Surf._build(self, X, Y, Z, C, C)

    @property
    def func(self):
        return self._func
    
    @func.setter
    @Mesh.param_setter
    def func(self, func):
        self._func = func

    @property
    def x_range(self):
        return self._x_range
    
    @x_range.setter
    @Mesh.param_setter
    def x_range(self, range):
        self._x_range = range

    @property
    def y_range(self):
        return self._y_range
    
    @y_range.setter
    @Mesh.param_setter
    def y_range(self, range):
        self._y_range = range

    @property
    def z_range(self):
        return self._z_range
    
    @z_range.setter
    @Mesh.param_setter
    def z_range(self, range):
        self._z_range = range

    @property
    def color_map(self):
        return self._color_map
    
    @color_map.setter
    @Mesh.param_setter
    def color_map(self, color_map:ColorMap):
        self._color_map = color_map
        self._use_color_map = True
        if not self._back_color_map_user_set:
            self._back_color_map = color_map
            self._back_use_color_map = True

    @property
    def back_color_map(self):
        return self._back_color_map
    
    @back_color_map.setter
    @Mesh.param_setter
    def back_color_map(self, color_map:ColorMap):
        self._back_color_map = color_map
        self._back_use_color_map = True
        self._back_color_map_user_set = True



import os

from OpenGL import GL
import pathlib
import warnings
import struct
import copy

from .Uniform import Uniform
from .GPUProgram import GPUProgram, LinkError, LinkWarning
from .Shaders import VertexShader, FragmentShader, GeometryShader, TessControlShader, TessEvaluationShader
from .Vertices import Vertices
from .Indices import Indices
from .Instances import Instances
from .utils import checktype, subscript, md5s, modify_time, save_var, load_var
from .GLInfo import GLInfo
from .GLConfig import GLConfig
from .GlassConfig import GlassConfig
from .TextureUnits import TextureUnits
from .ImageUnits import ImageUnits

class ShaderProgram(GPUProgram):

    __accum_draw_calls:dict = {}
    __accum_draw_points:dict = {}
    __accum_draw_lines:dict = {}
    __accum_draw_meshes:dict = {}
    __accum_draw_patches:dict = {}

    def __init__(self):
        GPUProgram.__init__(self)
        self.vertex_shader = VertexShader()
        self.fragment_shader = FragmentShader()
        self.geometry_shader = GeometryShader()
        self.tess_ctrl_shader = TessControlShader()
        self.tess_eval_shader = TessEvaluationShader()
        self._binary_file_name = ""
        self._meta_file_name = ""
        self._patch_vertices = 0
        self._context = 0
        self._include_paths = []

    @staticmethod
    def __update_dict(dest_dict, src_dict):
        for key, value in src_dict.items():
            dest_dict[key] = copy.copy(value)

    @checktype
    def add_include_path(self, include_path:str):
        self._include_paths.insert(0, include_path)

    @checktype
    def compile(self, file_name:str, shader_type:GLInfo.shader_types=None):
        if self._is_linked:
            raise RuntimeError("linked shader program cannot compile other shaders")
        
        if not os.path.isfile(file_name):
            raise FileNotFoundError(file_name)

        if shader_type is None:
            extention = pathlib.Path(file_name).suffix.lower()
            if extention in GLInfo.shader_ext_map:
                shader_type = GLInfo.shader_ext_map[extention]
            else:
                raise RuntimeError("Not support file type " + extention)

        shader = None
        if shader_type == GL.GL_VERTEX_SHADER:
            shader = VertexShader.load(file_name, include_paths=self._include_paths)
            self.vertex_shader = shader
        elif shader_type == GL.GL_FRAGMENT_SHADER:
            shader = FragmentShader.load(file_name, include_paths=self._include_paths)
            self.fragment_shader = shader
        elif shader_type == GL.GL_GEOMETRY_SHADER:
            shader = GeometryShader.load(file_name, include_paths=self._include_paths)
            self.geometry_shader = shader
            if shader.geometry_in in GLInfo.primitive_type_map:
                self._acceptable_primitives = GLInfo.primitive_type_map[shader.geometry_in]
        elif shader_type == GL.GL_TESS_CONTROL_SHADER:
            shader = TessControlShader.load(file_name, include_paths=self._include_paths)
            self.tess_ctrl_shader = shader
        elif shader_type == GL.GL_TESS_EVALUATION_SHADER:
            shader = TessEvaluationShader.load(file_name, include_paths=self._include_paths)
            self.tess_eval_shader = shader

        self._attributes_info.update(shader.attributes_info)
        self.__update_dict(self._uniforms_info, shader.uniforms_info)
        self.__update_dict(self._uniform_blocks_info, shader.uniform_blocks_info)
        self.__update_dict(self._shader_storage_blocks_info, shader.shader_storage_blocks_info)
        self._structs_info.update(shader.structs_info)
        self._outs_info.update(shader.outs_info)
        self._is_linked = False

    def _reapply(self):
        self.vertex_shader._apply()
        self.fragment_shader._apply()
        if self.geometry_shader.is_compiled:
            self.geometry_shader._apply()
        if self.tess_ctrl_shader.is_compiled:
            self.tess_ctrl_shader._apply()
        if self.tess_eval_shader.is_compiled:
            self.tess_eval_shader._apply()

        GL.glAttachShader(self._id, self.vertex_shader._id)
        if self.geometry_shader.is_compiled:
            GL.glAttachShader(self._id, self.geometry_shader._id)
        if self.tess_ctrl_shader.is_compiled:
            GL.glAttachShader(self._id, self.tess_ctrl_shader._id)
        if self.tess_eval_shader.is_compiled:
            GL.glAttachShader(self._id, self.tess_eval_shader._id)
        GL.glAttachShader(self._id, self.fragment_shader._id)

        related_files = "\n  " + "\n  ".join(self.related_files)
        if GlassConfig.print:
            print(f"linking program: {related_files}")

        GL.glProgramParameteri(self._id, GL.GL_PROGRAM_BINARY_RETRIEVABLE_HINT, GL.GL_TRUE)
        GL.glLinkProgram(self._id)
        
        message_bytes = GL.glGetProgramInfoLog(self._id)
        message = message_bytes
        if isinstance(message_bytes, bytes):
            message = str(message_bytes, encoding="utf-8")

        error_messages, warning_messages = self._format_error_warning(message)
        if warning_messages and GlassConfig.warning:
            warning_message = f"Warning when linking following files:{related_files}\n" + \
                              "\n".join(warning_messages)
            warnings.warn(warning_message, category=LinkWarning)

        if error_messages:
            error_message = f"Error when linking following files:{related_files}\n" + \
                            "\n".join(error_messages)
            raise LinkError(error_message)
        
        status = GL.glGetProgramiv(self._id, GL.GL_LINK_STATUS)
        if status != GL.GL_TRUE:
            raise LinkError(message)
            
        binary_length = int(GL.glGetProgramiv(self._id, GL.GL_PROGRAM_BINARY_LENGTH))
        length = GL.GLsizei()
        binary_format = GL.GLenum()
        binary_data = bytearray(binary_length)
        GL.glGetProgramBinary(self._id, binary_length, length, binary_format, binary_data)
        out_file = open(self._binary_file_name, "wb")
        out_file.write(struct.pack('i', int(binary_format.value)))
        out_file.write(struct.pack('i', binary_length))
        out_file.write(binary_data)
        out_file.close()

        self._apply_uniform_blocks()
        self._apply_shader_storage_blocks()

        meta_info = {}
        meta_info["attributes_info"] = self._attributes_info
        meta_info["acceptable_primitives"] = self._acceptable_primitives
        meta_info["uniforms_info"] = self._uniforms_info
        meta_info["uniform_blocks_info"] = self._uniform_blocks_info
        meta_info["shader_storage_blocks_info"] = self._shader_storage_blocks_info
        meta_info["structs_info"] = self._structs_info
        meta_info["outs_info"] = self._outs_info
        meta_info["sampler_map"] = self._sampler_map
        meta_info["uniform_map"] = self._uniform_map
        meta_info["uniform_block_map"] = self._uniform_block_map
        meta_info["shader_storage_block_map"] = self._shader_storage_block_map
        meta_info["include_paths"] = self._include_paths
        save_var(meta_info, self._meta_file_name)
        if GlassConfig.print:
            print("done")

    def _apply(self):
        if not self._linked_but_not_applied:
            return
        
        if self._id == 0:
            self._id = GL.glCreateProgram()
            if self._id == 0:
                raise MemoryError("failed to create ShaderProgram")

        if not self._should_relink:
            in_file = open(self._binary_file_name, "rb")
            binary_format = struct.unpack('i', in_file.read(4))[0]
            binary_length = struct.unpack('i', in_file.read(4))[0]
            binary_data = in_file.read(binary_length)
            in_file.close()

            status = GL.GL_FALSE
            try:
                GL.glProgramBinary(self._id, binary_format, binary_data, binary_length)
                status = GL.glGetProgramiv(self._id, GL.GL_LINK_STATUS)
            except:
                status = GL.GL_FALSE
                
            if GL.GL_TRUE != status:
                self._reapply()
        else:
            self._reapply()

        self._linked_but_not_applied = False

    def _test_should_relink(self):
        max_modify_time = 0
        shader_should_recompile = False

        binary_name = os.path.basename(self.vertex_shader.file_name)
        abs_file_names = os.path.abspath(self.vertex_shader.file_name).replace("\\", "/")
        shader_should_recompile = shader_should_recompile or self.vertex_shader._should_recompile
        if self.vertex_shader._max_modify_time > max_modify_time:
            max_modify_time = self.vertex_shader._max_modify_time

        if self.geometry_shader.is_compiled:
            binary_name += ("+" + os.path.basename(self.geometry_shader.file_name))
            abs_file_names += ("+" + os.path.abspath(self.geometry_shader.file_name).replace("\\", "/"))
            shader_should_recompile = shader_should_recompile or self.geometry_shader._should_recompile
            if self.geometry_shader._max_modify_time > max_modify_time:
                max_modify_time = self.geometry_shader._max_modify_time

        if self.tess_ctrl_shader.is_compiled:
            binary_name += ("+" + os.path.basename(self.tess_ctrl_shader.file_name))
            abs_file_names += ("+" + os.path.abspath(self.tess_ctrl_shader.file_name).replace("\\", "/"))
            shader_should_recompile = shader_should_recompile or self.tess_ctrl_shader._should_recompile
            if self.tess_ctrl_shader._max_modify_time > max_modify_time:
                max_modify_time = self.tess_ctrl_shader._max_modify_time

        if self.tess_eval_shader.is_compiled:
            binary_name += ("+" + os.path.basename(self.tess_eval_shader.file_name))
            abs_file_names += ("+" + os.path.abspath(self.tess_eval_shader.file_name).replace("\\", "/"))
            shader_should_recompile = shader_should_recompile or self.tess_eval_shader._should_recompile
            if self.tess_eval_shader._max_modify_time > max_modify_time:
                max_modify_time = self.tess_eval_shader._max_modify_time

        binary_name += ("+" + os.path.basename(self.fragment_shader.file_name))
        abs_file_names += ("+" + os.path.abspath(self.fragment_shader.file_name).replace("\\", "/"))
        shader_should_recompile = shader_should_recompile or self.fragment_shader._should_recompile
        if self.fragment_shader._max_modify_time > max_modify_time:
            max_modify_time = self.fragment_shader._max_modify_time

        base = GlassConfig.cache_folder + "/" + binary_name + "_" + md5s(GLConfig.renderer + "/" + abs_file_names)
        self._binary_file_name = base + ".bin"
        self._meta_file_name = base + ".meta"

        bin_mtime = modify_time(self._binary_file_name)
        meta_mtime = modify_time(self._meta_file_name)
        if not shader_should_recompile and \
           bin_mtime > 0 and meta_mtime > 0 and \
           max_modify_time < bin_mtime and \
           max_modify_time < meta_mtime:
            meta_info = load_var(self._meta_file_name)
            self._attributes_info = meta_info["attributes_info"]
            self._acceptable_primitives= meta_info["acceptable_primitives"]
            self._uniforms_info = meta_info["uniforms_info"]
            self._uniform_blocks_info = meta_info["uniform_blocks_info"]
            self._shader_storage_blocks_info = meta_info["shader_storage_blocks_info"]
            self._structs_info = meta_info["structs_info"]
            self._outs_info = meta_info["outs_info"]
            self._sampler_map = meta_info["sampler_map"]
            self._uniform_map = meta_info["uniform_map"]
            self._uniform_block_map = meta_info["uniform_block_map"]
            self._shader_storage_block_map = meta_info["shader_storage_block_map"]
            self._include_paths = meta_info["include_paths"]
            self._should_relink = False
        else:
            self._should_relink = True

        return self._should_relink

    def _link(self):
        if self._is_linked:
            return

        if not self.vertex_shader.is_compiled:
            raise RuntimeError("should compile vertex shader before link")

        if not self.fragment_shader.is_compiled:
            raise RuntimeError("should compile fragment shader before link")

        if self._test_should_relink():
            self._resolve_uniforms()
            self._resolve_uniform_blocks()
            self._resolve_shader_storage_blocks()

            keys = list(self._attributes_info.keys())
            for key in keys:
                if "location" not in self._attributes_info[key]:
                    location = GL.glGetAttribLocation(self._id, self._attributes_info[key]["name"])
                    self._attributes_info[key]["location"] = location
                    self._attributes_info[location] = self._attributes_info[key]

        self._is_linked = True
        self._linked_but_not_applied = True
        if GL.glCreateProgram:
            self._apply()

    def __check_vertices(self, vertices, start_index, total):
        len_vertices = len(vertices)
        if not GlassConfig.debug:
            if vertices is not None and total is None:
                total = len_vertices - start_index
            return total

        if vertices is not None and vertices:
            if start_index < 0 or (start_index >= len_vertices and vertices):
                raise IndexError("start index is out of range [0, " + str(len_vertices) + "]")

            if total is None:
                total = len_vertices - start_index

            if total < 0:
                raise IndexError("total number should be positive")

            if start_index + total > len_vertices:
                raise IndexError("end index is larger than array size")

        if not vertices:
            total = 0

        return total
    
    def __check_indices(self, indices, total):
        three_len_indices = 3 * len(indices)
        if not GlassConfig.debug:
            if indices is not None and total is None:
                total = three_len_indices
            return total

        if indices is not None and three_len_indices > 0:
            if total is None:
                total = three_len_indices

            if total < 0:
                raise IndexError("total number should be positive")

            if total > three_len_indices:
                raise IndexError("end index is larger than elements size")

        if three_len_indices == 0:
            total = 0

        return total

    def __update_uniforms(self):
        for uniform_var in self._uniform._uniform_var_map.values():
            var = uniform_var._bound_var
            if var is None:
                continue
            
            for atom_name, atom_info in Uniform._bound_vars[id(var)].items():
                atom_value = subscript(var, atom_info["subscript_chain"])
                self._uniform._set_atom(atom_name, atom_value)

        for atom_name, atom_value in self._uniform._should_update_atoms.items():
            self._uniform._set_atom(atom_name, atom_value)

        self._uniform._should_update_atoms.clear()

    def __update_samplers(self):
        self.use()
        self_used_texture_units = set()
        self_used_image_units = set()
        not_set_samplers = []
        not_set_images = []
        for sampler_info in self._sampler_map.values():
            location = sampler_info["location"]
            if location < 0:
                continue

            target_type = sampler_info["target_type"]
            texture_id = 0
            if sampler_info["sampler"] is not None:
                texture_id = sampler_info["sampler"].id

            texture_unit = None
            if "access" not in sampler_info:
                texture_unit = TextureUnits.unit_of_texture((target_type, texture_id))
                if texture_unit is None:
                    texture_unit = TextureUnits.available_unit
                    if texture_unit is None:
                        not_set_samplers.append(sampler_info)
                        continue
                    GLConfig.active_texture_unit = texture_unit
                    GL.glBindTexture(target_type, texture_id)
                    TextureUnits[texture_unit] = (target_type, texture_id)
                self_used_texture_units.add(texture_unit)
            else:
                texture_unit = ImageUnits.unit_of_image((target_type, texture_id))
                if texture_unit is None:
                    texture_unit = ImageUnits.available_unit
                    if texture_unit is None:
                        not_set_images.append(sampler_info)
                        continue
                
                access = sampler_info["access"]
                internal_format = sampler_info["sampler"].internal_format
                GL.glBindImageTexture(texture_unit, texture_id, 0, GL.GL_FALSE, 0, access, internal_format)
                ImageUnits[texture_unit] = (target_type, texture_id)
                self_used_image_units.add(texture_unit)

            if location not in self.uniform._texture_value_map or \
            self.uniform._texture_value_map[location] != texture_unit:
                GL.glUniform1i(location, texture_unit)
                self.uniform._texture_value_map[location] = texture_unit

        if not_set_samplers:
            available_units = GLConfig.available_texture_units - self_used_texture_units
            it_available_units = iter(available_units)
            for sampler_info in not_set_samplers:
                location = sampler_info["location"]

                texture_id = 0
                if sampler_info["sampler"] is not None:
                    texture_id = sampler_info["sampler"].id
                
                texture_unit = TextureUnits.unit_of_texture((target_type, texture_id))
                if texture_unit is None:
                    try:
                        texture_unit = next(it_available_units)
                    except:
                        raise RuntimeError(f"run out all {GLConfig.max_texture_units} texture units")
                
                target_type = sampler_info["target_type"]
                GLConfig.active_texture_unit = texture_unit
                GL.glBindTexture(target_type, texture_id)
                TextureUnits[texture_unit] = (target_type, texture_id)

                if location not in self.uniform._texture_value_map or \
                   self.uniform._texture_value_map[location] != texture_unit:
                    GL.glUniform1i(location, texture_unit)
                    self.uniform._texture_value_map[location] = texture_unit

        if not_set_images:
            available_units = GLConfig.available_image_units - self_used_image_units
            it_available_units = iter(available_units)
            for sampler_info in not_set_images:
                texture_id = 0
                if sampler_info["sampler"] is not None:
                    texture_id = sampler_info["sampler"].id

                texture_unit = ImageUnits.unit_of_image((target_type, texture_id))
                try:
                    texture_unit = next(it_available_units)
                except:
                    raise RuntimeError(f"run out all {GLConfig.max_image_units} image units")
                target_type = sampler_info["target_type"]
                access = sampler_info["access"]
                internal_format = sampler_info["sampler"].internal_format
                GL.glBindImageTexture(texture_unit, texture_id, 0, False, 0, access, internal_format)
                ImageUnits[texture_unit] = (target_type, texture_id)

                if location not in self.uniform._texture_value_map or \
                   self.uniform._texture_value_map[location] != texture_unit:
                    GL.glUniform1i(location, texture_unit)
                    self.uniform._texture_value_map[location] = texture_unit

    def __check_not_set_uniforms(self):
        if not self._uniform_not_set_warning or \
           not GlassConfig.debug or \
           not GlassConfig.warning:
            return

        not_set_uniforms = []
        for name, uniform_info in self._uniform_map.items():
            if not uniform_info["atoms"] and "location" not in uniform_info and \
               name not in self._uniform._atom_value_map:
                location = GL.glGetUniformLocation(self._id, name)
                uniform_info["location"] = location
                if location != -1:
                    not_set_uniforms.append(name)
                    
        if not_set_uniforms:
            warning_message = "in shader program:\n  "
            warning_message += "\n  ".join(self.related_files)
            warning_message += "\n"
            if len(not_set_uniforms) == 1:
                warning_message += f"uniform variable '{not_set_uniforms[0]}' is not set but used"
            else:
                warning_message += "following uniform variables are not set but used:\n"
                warning_message += "\n".join(not_set_uniforms)
                
            warnings.warn(warning_message, category=RuntimeWarning)

    def __preprocess_before_draw(self, primitive_type, vertices, indices, instances, start_index, total, times, is_patch):
        if GlassConfig.debug:
            if is_patch:
                if not self.tess_ctrl_shader.is_compiled:
                    raise RuntimeError("only shader program that with a tessilation shader can use draw_patches")

                if self._patch_vertices == 0:
                    raise RuntimeError("patch_vertices is not set before call draw_patches")
            else:
                if self.tess_ctrl_shader.is_compiled:
                    raise RuntimeError("shader program with a tessilation shader can only use draw_patches")
        
            if primitive_type is not None and \
               self._acceptable_primitives and \
               primitive_type not in self._acceptable_primitives:
                raise RuntimeError(f"geometry shader {self.geometry_shader.file_name}\nonly accept: {self._acceptable_primitives}, but {primitive_type.__repr__()} was given")

        if indices is None:
            total = self.__check_vertices(vertices, start_index, total)
        else:
            total = self.__check_indices(indices, total)

        if times is None and instances is not None:
            times = len(instances) // instances.divisor

        self.use()
        self.__update_uniforms()
        self.__update_samplers()
        self.__check_not_set_uniforms()
        if self._uniform_block.auto_upload:
            self._uniform_block.upload()
        if self._shader_storage_block.auto_upload:
            self._shader_storage_block.upload()

        if vertices is not None:
            vertices._apply(self, instances)

        if indices is not None:
            indices._apply()
            if vertices is not None:
                current_context = GLConfig.buffered_current_context
                key = (current_context, self, instances)
                if key in vertices._vao_map:
                    vertices._vao_map[key].setEBO(indices.ebo)
                else:
                    total = 0

        return total, times

    @property
    def patch_vertices(self):
        return self._patch_vertices

    @patch_vertices.setter
    @checktype
    def patch_vertices(self, value:int):
        self._patch_vertices = value

    @staticmethod
    def __increase_draw_calls():
        current_context = GLConfig.buffered_current_context
        if current_context not in ShaderProgram.__accum_draw_calls:
            ShaderProgram.__accum_draw_calls[current_context] = 0

        ShaderProgram.__accum_draw_calls[current_context] += 1

    @staticmethod
    def __increase_draw_points():
        current_context = GLConfig.buffered_current_context
        if current_context not in ShaderProgram.__accum_draw_points:
            ShaderProgram.__accum_draw_points[current_context] = 0

        ShaderProgram.__accum_draw_points[current_context] += 1

    @staticmethod
    def __increase_draw_lines():
        current_context = GLConfig.buffered_current_context
        if current_context not in ShaderProgram.__accum_draw_lines:
            ShaderProgram.__accum_draw_lines[current_context] = 0

        ShaderProgram.__accum_draw_lines[current_context] += 1

    @staticmethod
    def __increase_draw_meshes():
        current_context = GLConfig.buffered_current_context
        if current_context not in ShaderProgram.__accum_draw_meshes:
            ShaderProgram.__accum_draw_meshes[current_context] = 0

        ShaderProgram.__accum_draw_meshes[current_context] += 1

    @staticmethod
    def __increase_draw_patches():
        current_context = GLConfig.buffered_current_context
        if current_context not in ShaderProgram.__accum_draw_patches:
            ShaderProgram.__accum_draw_patches[current_context] = 0

        ShaderProgram.__accum_draw_patches[current_context] += 1

    @staticmethod
    def accum_draw_calls():
        current_context = GLConfig.buffered_current_context
        if current_context not in ShaderProgram.__accum_draw_calls:
            return 0
        
        return ShaderProgram.__accum_draw_calls[current_context]
    
    @staticmethod
    def accum_draw_points():
        current_context = GLConfig.buffered_current_context
        if current_context not in ShaderProgram.__accum_draw_points:
            return 0
        
        return ShaderProgram.__accum_draw_points[current_context]
    
    @staticmethod
    def accum_draw_lines():
        current_context = GLConfig.buffered_current_context
        if current_context not in ShaderProgram.__accum_draw_lines:
            return 0
        
        return ShaderProgram.__accum_draw_lines[current_context]
    
    @staticmethod
    def accum_draw_meshes():
        current_context = GLConfig.buffered_current_context
        if current_context not in ShaderProgram.__accum_draw_meshes:
            return 0
        
        return ShaderProgram.__accum_draw_meshes[current_context]

    @staticmethod
    def accum_draw_patches():
        current_context = GLConfig.buffered_current_context
        if current_context not in ShaderProgram.__accum_draw_patches:
            return 0
        
        return ShaderProgram.__accum_draw_patches[current_context]

    def draw_patches(self,
          vertices:Vertices=None, indices:Indices=None, instances:Instances=None,
          start_index:int=0, total:int=None, times:int=None):

        total, times = self.__preprocess_before_draw(
            None, vertices, indices, instances,
            start_index, total, times, True)
        
        if (total is not None and total <= 0) or \
           (times is not None and times <= 0):
            return

        GL.glPatchParameteri(GL.GL_PATCH_VERTICES, self._patch_vertices)

        self.use()
        if indices is not None:
            if times is None:
                GL.glDrawElements(GL.GL_PATCHES, total, GL.GL_UNSIGNED_INT, None)
            else:
                GL.glDrawElementsInstanced(GL.GL_PATCHES, total, GL.GL_UNSIGNED_INT, None, times)
        else:
            if total is None:
                total = self._patch_vertices
            if times is None:
                GL.glDrawArrays(GL.GL_PATCHES, start_index, total)
            else:
                GL.glDrawArraysInstanced(GL.GL_PATCHES, start_index, total, times)

        self.__increase_draw_patches()
        self.__increase_draw_calls()

    def draw_triangles(self,
            vertices:Vertices=None, indices:Indices=None, instances:Instances=None,
            primitive_type:GLInfo.triangle_types=GL.GL_TRIANGLES,
            start_index:int=0, total:int=None, times:int=None):

        total, times = self.__preprocess_before_draw(
            primitive_type, vertices, indices, instances,
            start_index, total, times, False)
        
        if (total is not None and total <= 0) or \
           (times is not None and times <= 0):
            return
        
        self.use()
        if indices is not None:
            if times is None:
                GL.glDrawElements(primitive_type, total, GL.GL_UNSIGNED_INT, None)
            else:
                GL.glDrawElementsInstanced(primitive_type, total, GL.GL_UNSIGNED_INT, None, times)
        else:
            if times is None:
                GL.glDrawArrays(primitive_type, start_index, total)
            else:
                GL.glDrawArraysInstanced(primitive_type, start_index, total, times)

        self.__increase_draw_meshes()
        self.__increase_draw_calls()

    def draw_points(self,
        vertices:Vertices=None, instances:Instances=None,
        start_index:int=0, total:int=None, times:int=None):

        total, times = self.__preprocess_before_draw(
            GL.GL_POINTS, vertices, None, instances,
            start_index, total, times, False)
        
        if (total is not None and total <= 0) or \
           (times is not None and times <= 0):
            return
        
        GL.glEnable(GL.GL_VERTEX_PROGRAM_POINT_SIZE)

        self.use()
        if times is None:
            GL.glDrawArrays(GL.GL_POINTS, start_index, total)
        else:
            GL.glDrawArraysInstanced(GL.GL_POINTS, start_index, total, times)

        self.__increase_draw_points()
        self.__increase_draw_calls()

    def draw_lines(self,
        vertices:Vertices=None, indices:Indices=None, instances:Instances=None,
        primitive_type:GLInfo.line_types=GL.GL_LINE_STRIP,
        start_index:int=0, total:int=None, times:int=None):

        total, times = self.__preprocess_before_draw(
            primitive_type, vertices, indices, instances,
            start_index, total, times, False)
        
        if (total is not None and total <= 0) or \
           (times is not None and times <= 0):
            return
    
        self.use()
        if indices is not None:
            if times is None:
                GL.glDrawElements(primitive_type, total, GL.GL_UNSIGNED_INT, None)
            else:
                GL.glDrawElementsInstanced(primitive_type, total, GL.GL_UNSIGNED_INT, None, times)
        else:
            if times is None:
                GL.glDrawArrays(primitive_type, start_index, total)
            else:
                GL.glDrawArraysInstanced(primitive_type, start_index, total, times)

        self.__increase_draw_lines()
        self.__increase_draw_calls()

    @property
    def related_files(self):
        result = []

        if self.vertex_shader._file_name:
            result.append(self.vertex_shader._file_name)
        if self.tess_ctrl_shader._file_name:
            result.append(self.tess_ctrl_shader._file_name)
        if self.tess_eval_shader._file_name:
            result.append(self.tess_eval_shader._file_name)
        if self.geometry_shader._file_name:
            result.append(self.geometry_shader._file_name)
        if self.fragment_shader._file_name:
            result.append(self.fragment_shader._file_name)

        return result



import re
from OpenGL import GL
from .utils import has_valid
from .helper import sizeof, type_from_str
from .GLInfo import GLInfo

class ShaderParser:

	@staticmethod
	def lines(content):
		return len(content.split("\n"))

	@staticmethod
	def line_of(content, pos):
		content_size = len(content)
		size = min(content_size, pos+1)
		n_line = 1
		for i in range(size):
			if content[i] == '\n':
				n_line += 1

		return n_line

	@staticmethod
	def skip_space(content, i):
		len_content = len(content)
		if i < 0 or i >= len_content:
			return i

		while i < len_content and content[i] in ' \t\n':
			i += 1

		return i

	@staticmethod
	def skip_space_reverse(content, i):
		if i < 0 or i >= len(content):
			return i

		while i >= 0 and content[i] in ' \t\n':
			i -= 1

		return i

	@staticmethod
	def join_name(name_list):
		result = ""
		for name in name_list:
			if isinstance(name, str):
				result += ("." + name)
			elif isinstance(name, int):
				result += ("[" + str(name) + "]")
		return result

	@staticmethod
	def delete_C_comments(content):
		result = ""
		n = len(content)
		in_str = False
		last_is_slash = False
		pos_start = 0
		i = 0
		while i < n:
			if content[i] == '"':
				last_is_slash = False
				if i-1 >= 0 and content[i-1] != '\\':
					in_str = not in_str
			elif content[i] == '/':
				if in_str:
					last_is_slash = False
					i += 1
					continue

				if last_is_slash:
					result += content[pos_start : i-1]
					last_is_slash = False
					i = content.find('\n', i)
					if i == -1:
						return result
					pos_start = i
				else:
					last_is_slash = True
			elif content[i] == '*':
				if in_str:
					last_is_slash = False
					i += 1
					continue

				if last_is_slash:
					result += content[pos_start : i-1]
					i += 1
					if i >= n:
						return result
					i = content.find("*/", i)
					if i == -1:
						return result
					i += 1
					pos_start = i + 1
				last_is_slash = False
				i += 1
				continue
			else:
				last_is_slash = False
			i += 1

		if pos_start < n:
			result += content[pos_start:]

		return re.sub(r"\\[ \t]*\n", "", result)

	@staticmethod
	def find_geometry_in(content):
		geometry_in = []
		regx = r"^\s*layout\s*\((?P<layout_qualifiers>[^\n]*?)\)\s*in\s*;"
		def append_geometry_in(match):
			layout_qualifiers = match.group("layout_qualifiers")
			args, kwargs = ShaderParser.get_layout_qualifiers(layout_qualifiers)
			acceptable_geometry_ins = ["points", "lines", "lines_adjacency", "triangles", "triangles_adjacency"]
			for arg in args:
				if arg in acceptable_geometry_ins:
					geometry_in.append(arg)
					return

		re.sub(regx, append_geometry_in, content, flags=re.M)
		
		if geometry_in:
			return geometry_in[0]
		else:
			return ""
		
	@staticmethod
	def find_attributes(content):
		attributes_info = {}
		def append_attribute(match):
			location = int(match.group("location"))
			var_type = match.group("type")
			python_type = type_from_str(var_type)
			size = sizeof(python_type)
			var_name = match.group("name")
			
			element_names = ShaderParser.resolve_array(var_name)
			for name in element_names:
				attribute = \
				{
					"name": name,
					"location": location,
					"type": var_type,
					"python_type": python_type,
					"size": size,
					"is_legacy": False
				}

				attributes_info[name] = attribute
				attributes_info[location] = attribute
				location += 1

			return ""

		def append_legacy_attribute(match):
			var_name = match.group("name")
			var_type = match.group("type")
			python_type = type_from_str(var_type)
			size = sizeof(python_type)

			if "[" in var_name:
				raise TypeError("not support attribute as array with legacy attribute format")
			
			attribute = \
			{
				"name": var_name,
				"type": var_type,
				"python_type": python_type,
				"size": size,
				"is_legacy": True
			}

			attributes_info[var_name] = attribute

			return ""

		re.sub(r"^\s*layout\s*\(\s*location\s*=\s*(?P<location>\d+)\)[\s\n]*in\s+(?P<type>[a-zA-Z_]\w*)\s+(?P<name>[a-zA-Z_]\w*(\s*\[\S+\]\s*)*)\s*;", append_attribute, content, flags=re.M)
		re.sub(r"^\s*attribute\s+(?P<type>[a-zA-Z_]\w*)\s+(?P<name>[a-zA-Z_]\w*(\s*\[\S+\]\s*)*)\s*;", append_legacy_attribute, content, flags=re.M)
		return attributes_info

	@staticmethod
	def resolve_name_type(var_name, var_type):
		pos_start = var_name.find('[')
		if pos_start == -1:
			return var_name, var_type
		var_type += var_name[pos_start:]
		return var_name[:pos_start].strip(" \t"), var_type

	@staticmethod
	def next_index(current_index, indices):
		i = -1
		try:
			while True:
				current_index[i] += 1
				if indices[i] != 0 and current_index[i] >= indices[i]:
					current_index[i] = 0
					i -= 1
				else:
					break
			return True
		except:
			return False

	@staticmethod
	def extract_array_indices(var_name):
		if "[" not in var_name:
			return []

		indices = []
		len_var_name = len(var_name)
		pos_start = len_var_name
		while True:
			pos_end = var_name.rfind("]", 0, pos_start)
			if pos_end == -1:
				break

			str_mid = var_name[pos_end+1:pos_start]
			if pos_end>=0 and pos_end < len_var_name and \
			   pos_start>=0 and pos_start < len_var_name and has_valid(str_mid):
				break

			pos_start = var_name.rfind("[", 0, pos_end)
			if pos_start == -1:
				break

			index = 0
			if pos_end - pos_start > 1:
				index = int(var_name[pos_start+1:pos_end].strip(" \t"))

			indices.insert(0, index)

		return indices

	@staticmethod
	def array_basename(name):
		pos_start = name.find("[")
		if pos_start == -1:
			return name
		else:
			return name[:pos_start].strip(" \t")

	@staticmethod
	def index_offset(total_index_str, current_index_str):
		total_index = ShaderParser.extract_array_indices(total_index_str)
		stop_index = ShaderParser.extract_array_indices(current_index_str)
		len_total_index = len(total_index)
		if len(stop_index) != len_total_index:
			raise ValueError("array should have same dimension")

		if not total_index:
			return 0

		current_index = [0]*len_total_index
		current_index[-1] = -1

		offset = 0
		while ShaderParser.next_index(current_index, total_index):
			if current_index == stop_index:
				break
			offset += 1

		return offset

	@staticmethod
	def resolve_array(var_name):
		if '[' not in var_name:
			return [var_name]
			
		indices = ShaderParser.extract_array_indices(var_name)
		base_name = ShaderParser.array_basename(var_name)

		element_names = []
		current_index = [0]*len(indices)
		current_index[-1] = -1
		while ShaderParser.next_index(current_index, indices):
			element_name = base_name
			for index in current_index:
				element_name += ("[" + str(index) + "]")
			element_names.append(element_name)

		return element_names

	@staticmethod
	def find_structs(content):
		structs_info = {}
		def append_struct(match):
			members = {}
			body = match.group("body")
			def append_member(match):
				var_name, var_type = ShaderParser.resolve_name_type(match.group("name"), match.group("type"))
				member = \
				{
					"type": var_type,
					"name": var_name
				}
				members[var_name] = member
				return ""

			re.sub(r"^\s*(?P<type>[a-zA-Z_]\w*(\s*\[\S+\]\s*)*)[\s\n]+(?P<name>[a-zA-Z_]\w*(\s*\[\S+\]\s*)*)[\s\n]*;", append_member, body, flags=re.M)
			structs_info[match.group("name")] = {"name": match.group("name"), "members": members}

		re.sub(r"^\s*struct[\s\n]+(?P<name>[a-zA-Z_]\w*)[\s\n]*\{(?P<body>.*?)\}", append_struct, content, flags=re.M|re.S)
		return structs_info

	@staticmethod
	def get_layout_qualifiers(content):
		if content is None:
			return [], {}

		items = content.split(",")
		args = []
		kwargs = {}
		for item in items:
			item = item.strip()
			if "=" in item:
				key_value = item.split("=")
				key = key_value[0].strip()
				value = key_value[1].strip()
				kwargs[key] = value
			else:
				args.append(item)

		return args, kwargs

	@staticmethod
	def get_memory_qualifiers(content):
		if content is None:
			return []

		result = []
		if "coherent" in content:
			result.append("coherent")

		if "volatile" in content:
			result.append("volatile")

		if "restrict" in content:
			result.append("restrict")

		if "readonly" in content:
			result.append("readonly")

		if "writeonly" in content:
			result.append("writeonly")

		if "readwrite" in content:
			result.append("readwrite")

		return result

	@staticmethod
	def get_internal_format(layout_args):
		for memory_qualifier, internel_format in GLInfo.memory_modifiers_to_internal_types_map.items():
			if memory_qualifier in layout_args:
				return internel_format
			
		return GL.GL_RGBA32F

	@staticmethod
	def find_uniforms(content):
		uniforms_info = {}
		atomic_offset = {}
		def append_uniform(match):
			var_name, var_type = ShaderParser.resolve_name_type(match.group("name"), match.group("type"))
			
			layout_args, layout_kwargs = [], {}
			internal_format = GL.GL_RGBA32F
			try:
				layout_qualifiers_str = match.group("layout_qualifiers")
				layout_args, layout_kwargs = ShaderParser.get_layout_qualifiers(layout_qualifiers_str)
				internal_format = ShaderParser.get_internal_format(layout_args)
			except IndexError:
				pass

			binding_point = -1
			if "binding" in layout_kwargs:
				binding_point = int(layout_kwargs["binding"])

			offset = 0
			if "offset" in layout_kwargs:
				offset = int(layout_kwargs["offset"])

			if binding_point != -1 and var_type == "atomic_uint":
				if offset == 0 and binding_point in atomic_offset:
					offset = atomic_offset[binding_point] + 4
				atomic_offset[binding_point] = offset

			memory_qualifiers = []
			try:
				memory_qualifiers_str = match.group("memory_qualifiers")
				memory_qualifiers = ShaderParser.get_memory_qualifiers(memory_qualifiers_str)
			except IndexError:
				pass

			uniforms_info[var_name] = \
			{
				"name": var_name,
				"type": var_type,
				"subscript_chain": [],
				"layout_qualifiers": (layout_args, layout_kwargs),
				"memory_qualifiers": memory_qualifiers,
				"internal_format": internal_format,
				"binding_point": binding_point,
				"offset": offset
			}
			return ""

		uniform_prefix = r"uniform[\s\n]+"
		uniform_regx = r"(?P<type>[a-zA-Z_]\w*(\s*\[\S+\]\s*)*)[\s\n]+(?P<name>[a-zA-Z_]\w*(\s*\[\S+\]\s*)*)[\s\n]*(=.*?)?;";
		layout_qualifiers_regx = r"(layout\s*\((?P<layout_qualifiers>[^\n]*?)\)[\s\n]*)?"
		memory_qualifiers_regx = r"(?P<memory_qualifiers>((coherent|volatile|restrict|readonly|writeonly|readwrite)[\s\n]+)+)?"
		re.sub(r"^\s*" + layout_qualifiers_regx + memory_qualifiers_regx + uniform_prefix + uniform_regx, append_uniform, content, flags=re.M|re.S)
		re.sub(r"^\s*" + uniform_prefix + layout_qualifiers_regx + memory_qualifiers_regx + uniform_regx, append_uniform, content, flags=re.M|re.S)
		return uniforms_info

	@staticmethod
	def find_uniform_blocks(content):
		uniform_blocks_info = {}
		def append_uniform_block(match):
			members = {}
			body = match.group("body")
			def append_member(match):
				var_name, var_type = ShaderParser.resolve_name_type(match.group("name"), match.group("type"))
				member = \
				{
					"type": var_type,
					"name": var_name
				}
				members[var_name] = member
				return ""

			re.sub(r"^\s*(?P<type>[a-zA-Z_]\w*(\s*\[\S+\]\s*)*)[\s\n]+(?P<name>[a-zA-Z_]\w*(\s*\[\S+\]\s*)*)[\s\n]*;", append_member, body, flags=re.M|re.S)
			var_name = ""
			try:
				var_name = match.group("var_name")
				if var_name is None:
					var_name = ""
			except:
				var_name = ""

			layout_args, layout_kwargs = [], {}
			internal_format = GL.GL_RGBA32F
			try:
				layout_qualifiers_str = match.group("layout_qualifiers")
				layout_args, layout_kwargs = ShaderParser.get_layout_qualifiers(layout_qualifiers_str)
				internal_format = ShaderParser.get_internal_format(layout_args)
			except IndexError:
				pass

			binding_point = -1
			if "binding" in layout_kwargs:
				binding_point = int(layout_kwargs["binding"])

			memory_qualifiers = []
			try:
				memory_qualifiers_str = match.group("memory_qualifiers")
				memory_qualifiers = ShaderParser.get_memory_qualifiers(memory_qualifiers_str)
			except IndexError:
				pass

			uniform_blocks_info[match.group("name")] = \
			{
				"name": match.group("name"),
				"members": members,
				"var_name": var_name,
				"layout_qualifiers": (layout_args, layout_kwargs),
				"memory_qualifiers": memory_qualifiers,
				"internal_format": internal_format,
				"binding_point": binding_point
			}
		
		uniform_prefix = r"uniform[\s\n]+"
		uniform_block_regx = r"(?P<name>[a-zA-Z_]\w*)[\s\n]*\{(?P<body>.*?)\}\s*((?P<var_name>[a-zA-Z_]\w*)\s*)?;"
		layout_qualifiers_regx = r"(layout\s*\((?P<layout_qualifiers>[^\n]*?)\)[\s\n]*)?"
		memory_qualifiers_regx = r"(?P<memory_qualifiers>((coherent|volatile|restrict|readonly|writeonly|readwrite)[\s\n]+)+)?"
		re.sub(r"^\s*" + layout_qualifiers_regx + memory_qualifiers_regx + uniform_prefix + uniform_block_regx, append_uniform_block, content, flags=re.M|re.S)
		re.sub(r"^\s*" + uniform_prefix + layout_qualifiers_regx + memory_qualifiers_regx + uniform_block_regx, append_uniform_block, content, flags=re.M|re.S)
		return uniform_blocks_info

	@staticmethod
	def find_shader_storage_blocks(content):
		shader_storage_blocks_info = {}
		def append_shader_storage_block(match):
			members = {}
			body = match.group("body")
			def append_member(match):
				var_name, var_type = ShaderParser.resolve_name_type(match.group("name"), match.group("type"))
				member = \
				{
					"type": var_type,
					"name": var_name
				}
				members[var_name] = member
				return ""

			re.sub(r"^\s*(?P<type>[a-zA-Z_]\w*(\s*\[.*\]\s*)*)[\s\n]+(?P<name>[a-zA-Z_]\w*(\s*\[.*\]\s*)*)[\s\n]*;", append_member, body, flags=re.M|re.S)
			var_name = ""
			try:
				var_name = match.group("var_name")
				if var_name is None:
					var_name = ""
			except:
				var_name = ""

			layout_args, layout_kwargs = [], {}
			internal_format = GL.GL_RGBA32F
			try:
				layout_qualifiers_str = match.group("layout_qualifiers")
				layout_args, layout_kwargs = ShaderParser.get_layout_qualifiers(layout_qualifiers_str)
				internal_format = ShaderParser.get_internal_format(layout_args)
			except IndexError:
				pass

			binding_point = -1
			if "binding" in layout_kwargs:
				binding_point = int(layout_kwargs["binding"])

			memory_qualifiers = []
			try:
				memory_qualifiers_str = match.group("memory_qualifiers")
				memory_qualifiers = ShaderParser.get_memory_qualifiers(memory_qualifiers_str)
			except IndexError:
				pass
			
			shader_storage_blocks_info[match.group("name")] = \
			{
				"name": match.group("name"),
				"members": members,
				"var_name": var_name,
				"layout_qualifiers": (layout_args, layout_kwargs),
				"memory_qualifiers": memory_qualifiers,
				"internal_format": internal_format,
				"binding_point": binding_point
			}
		
		buffer_prefix = r"buffer[\s\n]+"
		buffer_regx = r"(?P<name>[a-zA-Z_]\w*)[\s\n]*\{(?P<body>.*?)\}\s*((?P<var_name>[a-zA-Z_]\w*)\s*)?;"
		layout_qualifiers_regx = r"(layout\s*\((?P<layout_qualifiers>[^\n]*?)\)[\s\n]*)?"
		memory_qualifiers_regx = r"(?P<memory_qualifiers>((coherent|volatile|restrict|readonly|writeonly|readwrite)[\s\n]+)+)?"
		re.sub(r"^\s*" + layout_qualifiers_regx + memory_qualifiers_regx + buffer_prefix + buffer_regx, append_shader_storage_block, content, flags=re.M|re.S)
		re.sub(r"^\s*" + buffer_prefix + layout_qualifiers_regx + memory_qualifiers_regx + buffer_regx, append_shader_storage_block, content, flags=re.M|re.S)
		return shader_storage_blocks_info

	@staticmethod
	def find_outs(content):
		outs_info = {}
		def append_out(match):
			var_name, var_type = ShaderParser.resolve_name_type(match.group("name"), match.group("type"))
			outs_info[var_name] = {"name": var_name, "type": var_type}
			try:
				outs_info[var_name]["location"] = int(match.group("location"))
			except:
				outs_info[var_name]["location"] = 0

			return ""

		re.sub(r"^(\s*layout\s*\(\s*location\s*=\s*(?P<location>\d+)\))*[\s\n]*out[\s\n]+(?P<type>[a-zA-Z_]\w*(\s*\[\S+\]\s*)*)[\s\n]+(?P<name>[a-zA-Z_]\w*(\s*\[\S+\]\s*)*)[\s\n]*(=.*?)?;", append_out, content, flags=re.M|re.S)
		return outs_info

	@staticmethod
	def find_work_group_size(content):
		group_size = []

		def get_size3(match):
			group_size.append(int(match.group("local_size_x")))
			group_size.append(int(match.group("local_size_y")))
			group_size.append(int(match.group("local_size_z")))
			return ""

		def get_size2(match):
			group_size.append(int(match.group("local_size_x")))
			group_size.append(int(match.group("local_size_y")))
			group_size.append(1)
			return ""

		def get_size1(match):
			group_size.append(match.group("local_size_x"))
			group_size.append(1)
			group_size.append(1)
			return ""

		re.sub(r"^\s*layout\s*\(\s*local_size_x\s*=\s*(?P<local_size_x>\d+)\s*,\s*local_size_y\s*=\s*(?P<local_size_y>\d+)\s*,\s*local_size_z\s*=\s*(?P<local_size_z>\d+)\s*\)\s*in\s*;", get_size3, content, flags=re.M|re.S)
		if not group_size:
			re.sub(r"^\s*layout\s*\(\s*local_size_x\s*=\s*(?P<local_size_x>\d+)\s*,\s*local_size_y\s*=\s*(?P<local_size_y>\d+)\s*\)\s*in\s*;", get_size2, content, flags=re.M|re.S)
		
		if not group_size:
			re.sub(r"^\s*layout\s*\(\s*local_size_x\s*=\s*(?P<local_size_x>\d+)\s*\)\s*in\s*;", get_size1, content, flags=re.M|re.S)

		return tuple(group_size)


from .VAO import VAO
from .VBO import VBO
from .EBO import EBO
from .FBO import FBO
from .RBO import RBO
from .SSBO import SSBO
from .UBO import UBO
from .ACBO import ACBO

from .sampler2D import sampler2D
from .usampler2D import usampler2D
from .isampler2D import isampler2D
from .image2D import image2D
from .uimage2D import uimage2D
from .iimage2D import iimage2D
from .sampler2DMS import sampler2DMS
from .isampler2DMS import isampler2DMS
from .usampler2DMS import usampler2DMS
from .samplerCube import samplerCube
from .sampler2DArray import sampler2DArray

from .Vertices import Vertices, Vertex
from .Indices import Indices
from .Instances import Instance, Instances
from .AttrList import AttrList

from .ShaderProgram import ShaderProgram
from .ComputeProgram import ComputeProgram

from .ShaderStorageBlock import ShaderStorageBlock
from .UniformBlock import UniformBlock
from .Uniform import Uniform

from .GLConfig import GLConfig
from .GlassConfig import GlassConfig
from .GLInfo import GLInfo
from .RenderHints import RenderHints
from .download import download


from .WeakRef import WeakRef

class WeakList:

    class iterator:
        def __init__(self, _iterator):
            self._iterator = _iterator

        def __next__(self):
            return self._iterator.__next__().value
        
        def __iter__(self):
            return self

    def __init__(self):
        self._list = []

    def __setitem__(self, index, value):
        self._list[index] = WeakRef(value)

    def __getitem__(self, index):
        value = self._list[index]
        if isinstance(index, slice):
            return list(map(lambda x:x.value, value))
        else:
            return value.value
            
    def __delitem__(self, index):
        del self._list[index]
        
    def __contains__(self, value):
        return WeakRef(value) in self._list
        
    def __len__(self):
        return self._list.__len__()
    
    def __bool__(self):
        return bool(self._list)
    
    def __iter__(self):
        return WeakList.iterator(self._list.__iter__())
    
    def append(self, value):
        self._list.append(WeakRef(value))

    def insert(self, index, value):
        self._list.insert(index, WeakRef(value))

    def extend(self, value_list):
        if isinstance(value_list, WeakList):
            self._list.extend(value_list._list)
        else:
            self._list.extend(map(lambda x:WeakRef(x), value_list))

    def remove(self, value):
        self._list.remove(WeakRef(value))

    def pop(self, index):
        return self._list.pop(index).value
        
    def clear(self):
        self._list.clear()
        


from .SkyBox import SkyBox
from .SkyDome import SkyDome

from glass import samplerCube, sampler2D

import glm

class Background:
    def __init__(self):
        self._skybox:SkyBox = SkyBox()
        self._skydome:SkyDome = SkyDome()
        self._color:glm.vec4 = glm.vec4(0)
        self._distance:float = 100.0

    @property
    def skybox(self):
        return self._skybox
    
    @skybox.setter
    def skybox(self, skybox_map:samplerCube):
        self._skybox.skybox_map = skybox_map
    
    @property
    def skydome(self):
        return self._skydome
    
    @skydome.setter
    def skydome(self, image:str):
        self._skydome.skydome_map = image

    @property
    def skybox_map(self)->samplerCube:
        return self._skybox.skybox_map
    
    @property
    def skydome_map(self)->sampler2D:
        return self._skydome.skydome_map

    @property
    def distance(self)->float:
        return self._distance
    
    @distance.setter
    def distance(self, distance:float)->None:
        self._distance = distance

    @property
    def color(self)->glm.vec4:
        return self._color
    
    @color.setter
    def color(self, color:(glm.vec4,glm.vec3))->None:
        if isinstance(color, glm.vec3):
            color = glm.vec4(color, 1)

        self._color = color


import setuptools
import os
import shutil
import sys
import platform
import subprocess
import warnings

# install PyGLM
if "-d" in sys.argv:
    plat = "x64" if platform.architecture()[0] == "64bit" else "win32"
    version = ".".join(platform.python_version().split(".")[:2])
    if int(version.split(".")[1]) < 12:
        subprocess.call([sys.executable, "-m", "pip", "install", "PyGLM"])
    else:
        temp_folder = os.path.dirname(os.path.abspath(__file__))

        target_file = ""
        gitee_url = ""
        github_url = ""
        
        if plat == "x64":
            gitee_url = "https://gitee.com/time-coder/Glass-Engine/raw/main/PyGLM/PyGLM-2.7.0-cp312-cp312-win_amd64.whl"
            github_url = "https://raw.githubusercontent.com/Time-Coder/Glass-Engine/main/PyGLM/PyGLM-2.7.0-cp312-cp312-win_amd64.whl"
            target_file = temp_folder + "/PyGLM-2.7.0-cp312-cp312-win_amd64.whl"
        else:
            gitee_url = "https://gitee.com/time-coder/Glass-Engine/raw/main/PyGLM/PyGLM-2.7.0-cp312-cp312-win32.whl"
            github_url = "https://raw.githubusercontent.com/Time-Coder/Glass-Engine/main/PyGLM/PyGLM-2.7.0-cp312-cp312-win32.whl"
            target_file = temp_folder + "/PyGLM-2.7.0-cp312-cp312-win32.whl"

        subprocess.call([sys.executable, "-m", "pip", "install", "wget"])
        import wget

        try:
            wget.download(gitee_url, target_file)
        except:
            wget.download(github_url, target_file)

        return_code = subprocess.call([sys.executable, "-m", "pip", "install", target_file])
        if return_code != 0:
            warnings.warn("Failed to install PyGLM, Please install it manually.")

def find_files(module, directory):
    file_list = []
    for root, _, files in os.walk(directory):
        for file in files:
            file_path = os.path.join(root, file)[(len(module) + 1) :]
            if "__glcache__" not in file_path:
                file_list.append(file_path.replace("\\", "/"))
    return file_list

if os.path.isdir("build"):
    shutil.rmtree("build")

if os.path.isdir("glass_engine.egg-info"):
    shutil.rmtree("glass_engine.egg-info")

with open("glass_engine/README_PYPI.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

glass_engine_extra_files = \
[
    "images/glass_engine_logo64.png",
    "images/glass_engine_logo256.png",
    "images/start.png",
    "README_PYPI.md"
]
glass_engine_extra_files.extend(find_files("glass_engine", "glass_engine/glsl"))
glass_extra_files = find_files("glass", "glass/glsl")

setuptools.setup(
    name="glass_engine",
    version="0.1.15",
    author="王炳辉 (BingHui-WANG)",
    author_email="binghui.wang@foxmail.com",
    description="An easy-to-use 3D rendering engine for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Time-Coder/Glass-Engine",
    packages=setuptools.find_packages(),
    package_data={
        'glass_engine': glass_engine_extra_files,
        'glass': glass_extra_files
    },
    platforms=["win_amd64", "win32"],
    python_requires=">=3.6",
    install_requires=[
        "PyOpenGL",
        "PyOpenGL_accelerate",
        "moderngl",
        "MarkupSafe==2.0.1",
        "qt-material",
        "numpy",
        "opencv-python",
        "Pillow",
        "maxminddb-geolite2",
        "wget",
        "requests",
        "chardet"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
    ],
)



from .PostProcessEffect import PostProcessEffect
from ..Frame import Frame
from glass.utils import checktype
from glass import ShaderProgram, FBO, sampler2D, sampler2DArray, samplerCube, GLConfig, ShaderStorageBlock

import numpy as np
from OpenGL import GL
import os

class KernelFilter(PostProcessEffect):

    class Kernel(ShaderStorageBlock.HostClass):
        def __init__(self, kernel:np.ndarray):
            ShaderStorageBlock.HostClass.__init__(self)
            
            self._rows = kernel.shape[0]
            self._cols = kernel.shape[1]
            self._data = list(kernel.flatten())
            self._kernel = kernel
        
        @property
        def rows(self):
            return self._rows
        
        @property
        def cols(self):
            return self._cols
        
        @property
        def data(self):
            return self._data
        
        @property
        def kernel(self):
            return self._kernel

        @kernel.setter
        @ShaderStorageBlock.HostClass.not_const
        def kernel(self, kernel:np.ndarray):
            self.rows = kernel.shape[0]
            self.cols = kernel.shape[1]
            self.data = list(kernel.flatten())
            self._kernel = kernel

    @checktype
    def __init__(self, kernel:np.ndarray):
        PostProcessEffect.__init__(self)
        self._kernel = KernelFilter.Kernel(kernel)

        self._program = None
        self._cube_program = None
        self._fbo = None
        self._array_fbo = None
        self._cube_fbo = None
        self.__array_programs = {}

    def need_pos_info(self) -> bool:
        return False

    @property
    def program(self):
        if self._program is None:
            self._program = ShaderProgram()
            self._program.compile(Frame.draw_frame_vs)
            self._program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/kernel_filter.fs")
            self._program["Kernel"].bind(self._kernel)
        return self._program

    @property
    def cube_program(self):
        if self._cube_program is None:
            self._cube_program = ShaderProgram()
            self._cube_program.compile(Frame.draw_frame_vs)
            self._cube_program.compile(Frame.draw_frame_array_gs(6))
            self._cube_program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/kernel_cube_filter.fs")
            self._cube_program["Kernel"].bind(self._kernel)
        return self._cube_program

    @property
    def fbo(self):
        if self._fbo is None:
            self._fbo = FBO()
            self._fbo.attach(0, sampler2D)
        return self._fbo

    @property
    def array_fbo(self):
        if self._array_fbo is None:
            self._array_fbo = FBO()
            self._array_fbo.attach(0, sampler2DArray)
        return self._array_fbo

    @property
    def cube_fbo(self):
        if self._cube_fbo is None:
            self._cube_fbo = FBO()
            self._cube_fbo.attach(0, samplerCube)
        return self._cube_fbo

    def __call__(self, screen_image:(sampler2D,sampler2DArray,samplerCube))->(sampler2D,sampler2DArray,samplerCube):
        if isinstance(screen_image, sampler2D):
            self.fbo.resize(screen_image.width, screen_image.height)
            with GLConfig.LocalConfig(depth_test=False, cull_face=None, polygon_mode=GL.GL_FILL):
                with self.fbo:
                    self.program["screen_image"] = screen_image
                    self.program.draw_triangles(Frame.vertices, Frame.indices)

            return self.fbo.color_attachment(0)
        elif isinstance(screen_image, sampler2DArray):
            self.array_fbo.resize(screen_image.width, screen_image.height, layers=screen_image.layers)
            program = self.array_program(screen_image.layers)
            with GLConfig.LocalConfig(depth_test=False, cull_face=None, polygon_mode=GL.GL_FILL):
                with self.array_fbo:
                    program["screen_image"] = screen_image
                    program.draw_triangles(Frame.vertices, Frame.indices)

            return self.array_fbo.color_attachment(0)
        elif isinstance(screen_image, samplerCube):
            self.cube_fbo.resize(screen_image.width, screen_image.height)
            with GLConfig.LocalConfig(depth_test=False, cull_face=None, polygon_mode=GL.GL_FILL):
                with self.cube_fbo:
                    self.cube_program["screen_image"] = screen_image
                    self.cube_program.draw_triangles(Frame.vertices, Frame.indices)

            return self.cube_fbo.color_attachment(0)
        
    def draw_to_active(self, screen_image: sampler2D) -> None:
        with GLConfig.LocalConfig(depth_test=False, cull_face=None, polygon_mode=GL.GL_FILL):
            self.program["screen_image"] = screen_image
            self.program.draw_triangles(Frame.vertices, Frame.indices)

    @property
    def kernel(self)->np.ndarray:
        return self._kernel.kernel
    
    @kernel.setter
    def kernel(self, kernel:np.ndarray):
        self._kernel.kernel = kernel
    
    def array_program(self, layers):
        if layers in self.__array_programs:
            return self.__array_programs[layers]

        program = ShaderProgram()
        program.compile(Frame.draw_frame_vs)
        program.compile(Frame.draw_frame_array_gs(layers))
        program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/kernel_array_filter.fs")
        program["Kernel"].bind(self._kernel)

        self.__array_programs[layers] = program

        return program
    


from .Scene import Scene
from .Camera import Camera
from .Geometries.Floor import Floor
from .Lights.DirLight import DirLight
from .Manipulators import ModelViewManipulator

def SceneRoam(add_floor=True):
    scene = Scene()

    camera = Camera()
    camera.position.y = -5
    camera.position.z = 1.7
    camera.pitch = -10
    scene.add(camera)

    floor = None
    if add_floor:
        floor = Floor()
        scene.add(floor)

    dir_light = DirLight()
    dir_light.pitch = -45
    dir_light.yaw = 45
    scene.add(dir_light)

    return scene, camera, dir_light, floor

def ModelView(distance:float=5, azimuth_deg:float=0, elevation_deg:float=0):
    scene = Scene()

    camera = Camera()
    camera.screen.manipulator = ModelViewManipulator(distance, azimuth_deg, elevation_deg)
    scene.add(camera)

    dir_light = DirLight()
    dir_light.pitch = -45
    dir_light.yaw = 45
    scene.add(dir_light)

    return scene, camera, dir_light